from .common import AliasesEnum, Count, MailingAddress, OrderCancellation, MoneyBag
from .utils import Extract, DeepExtract


class Orders(AliasesEnum):
    """A class to represent a Shopify GraphQL order.
    Reference: https://shopify.dev/docs/api/admin-graphql/latest/objects/Order
    Require `read_orders`, `read_marketplace_orders` or `read_quick_sale` permission.
    """

    # additionalFees = "additionalFees"
    # agreements = "agreements"
    # alerts = "alerts"
    # app = "app"
    # billingAddress = "billingAddress"
    billingAddressMatchesShippingAddress = "billingAddressMatchesShippingAddress"
    cancellation = OrderCancellation
    cancelledAt = "cancelledAt"
    cancelReason = "cancelReason"
    canMarkAsPaid = "canMarkAsPaid"
    canNotifyCustomer = "canNotifyCustomer"
    capturable = "capturable"
    # cartDiscountAmountSet = "cartDiscountAmountSet"
    # channelInformation = "channelInformation"
    clientIp = "clientIp"
    closed = "closed"
    closedAt = "closedAt"
    confirmationNumber = "confirmationNumber"
    confirmed = "confirmed"
    createdAt = "createdAt"
    currencyCode = "currencyCode"
    # currentCartDiscountAmountSet = "currentCartDiscountAmountSet"
    currentShippingPriceSet = MoneyBag
    currentShippingPriceSet_presentmentMoney_amount = DeepExtract(
        ["currentShippingPriceSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    currentShippingPriceSet_presentmentMoney_currencyCode = DeepExtract(
        ["currentShippingPriceSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    currentShippingPriceSet_shopMoney_amount = DeepExtract(
        ["currentShippingPriceSet", "shopMoney", "amount"], "DECIMAL"
    )
    currentShippingPriceSet_shopMoney_currencyCode = DeepExtract(
        ["currentShippingPriceSet", "shopMoney", "currencyCode"], "TEXT"
    )
    currentSubtotalLineItemsQuantity = "currentSubtotalLineItemsQuantity"
    currentSubtotalPriceSet = MoneyBag
    currentSubtotalPriceSet_presentmentMoney_amount = DeepExtract(
        ["currentSubtotalPriceSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    currentSubtotalPriceSet_presentmentMoney_currencyCode = DeepExtract(
        ["currentSubtotalPriceSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    currentSubtotalPriceSet_shopMoney_amount = DeepExtract(
        ["currentSubtotalPriceSet", "shopMoney", "amount"], "DECIMAL"
    )
    currentSubtotalPriceSet_shopMoney_currencyCode = DeepExtract(
        ["currentSubtotalPriceSet", "shopMoney", "currencyCode"], "TEXT"
    )
    # currentTaxLines = "currentTaxLines"
    currentTotalAdditionalFeesSet = MoneyBag
    currentTotalAdditionalFeesSet_presentmentMoney_amount = DeepExtract(
        ["currentTotalAdditionalFeesSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    currentTotalAdditionalFeesSet_presentmentMoney_currencyCode = DeepExtract(
        ["currentTotalAdditionalFeesSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    currentTotalAdditionalFeesSet_shopMoney_amount = DeepExtract(
        ["currentTotalAdditionalFeesSet", "shopMoney", "amount"], "DECIMAL"
    )
    currentTotalAdditionalFeesSet_shopMoney_currencyCode = DeepExtract(
        ["currentTotalAdditionalFeesSet", "shopMoney", "currencyCode"], "TEXT"
    )
    currentTotalDiscountsSet = MoneyBag
    currentTotalDiscountsSet_presentmentMoney_amount = DeepExtract(
        ["currentTotalDiscountsSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    currentTotalDiscountsSet_presentmentMoney_currencyCode = DeepExtract(
        ["currentTotalDiscountsSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    currentTotalDiscountsSet_shopMoney_amount = DeepExtract(
        ["currentTotalDiscountsSet", "shopMoney", "amount"], "DECIMAL"
    )
    currentTotalDiscountsSet_shopMoney_currencyCode = DeepExtract(
        ["currentTotalDiscountsSet", "shopMoney", "currencyCode"], "TEXT"
    )
    currentTotalDutiesSet = MoneyBag
    currentTotalDutiesSet_presentmentMoney_amount = DeepExtract(
        ["currentTotalDutiesSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    currentTotalDutiesSet_presentmentMoney_currencyCode = DeepExtract(
        ["currentTotalDutiesSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    currentTotalDutiesSet_shopMoney_amount = DeepExtract(["currentTotalDutiesSet", "shopMoney", "amount"], "DECIMAL")
    currentTotalDutiesSet_shopMoney_currencyCode = DeepExtract(
        ["currentTotalDutiesSet", "shopMoney", "currencyCode"], "TEXT"
    )
    currentTotalPriceSet = MoneyBag
    currentTotalPriceSet_presentmentMoney_amount = DeepExtract(
        ["currentTotalPriceSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    currentTotalPriceSet_presentmentMoney_currencyCode = DeepExtract(
        ["currentTotalPriceSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    currentTotalPriceSet_shopMoney_amount = DeepExtract(["currentTotalPriceSet", "shopMoney", "amount"], "DECIMAL")
    currentTotalPriceSet_shopMoney_currencyCode = DeepExtract(
        ["currentTotalPriceSet", "shopMoney", "currencyCode"], "TEXT"
    )
    currentTotalTaxSet = MoneyBag
    currentTotalTaxSet_presentmentMoney_amount = DeepExtract(
        ["currentTotalTaxSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    currentTotalTaxSet_presentmentMoney_currencyCode = DeepExtract(
        ["currentTotalTaxSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    currentTotalTaxSet_shopMoney_amount = DeepExtract(["currentTotalTaxSet", "shopMoney", "amount"], "DECIMAL")
    currentTotalTaxSet_shopMoney_currencyCode = DeepExtract(["currentTotalTaxSet", "shopMoney", "currencyCode"], "TEXT")
    currentTotalWeight = "currentTotalWeight"
    # customAttributes = "customAttributes"
    # customer = "customer"
    customerId = Extract("customer", "id")  # custom
    customerAcceptsMarketing = "customerAcceptsMarketing"
    # customerJourneySummary = "customerJourneySummary"
    customerLocale = "customerLocale"
    # discountApplications = "discountApplications"
    discountCode = "discountCode"
    discountCodes = "discountCodes"
    # displayAddress = "displayAddress"
    # displayFinancialStatus = "displayFinancialStatus"
    # displayFulfillmentStatus = "displayFulfillmentStatus"
    # disputes = "disputes"
    dutiesIncluded = "dutiesIncluded"
    edited = "edited"
    email = "email"
    estimatedTaxes = "estimatedTaxes"
    # events = "events"
    fulfillable = "fulfillable"
    # fulfillmentOrders = "fulfillmentOrders"
    # fulfillments = "fulfillments"
    fulfillmentsCount = Count
    fullyPaid = "fullyPaid"
    hasTimelineComment = "hasTimelineComment"
    id = "id"
    # legacyResourceId = "legacyResourceId"
    # lineItems = "lineItems"
    # localizedFields = "localizedFields"
    # merchantBusinessEntity = "merchantBusinessEntity"
    merchantEditable = "merchantEditable"
    merchantEditableErrors = "merchantEditableErrors"
    # merchantOfRecordApp = "merchantOfRecordApp"
    # metafield = "metafield"
    # metafields = "metafields"
    name = "name"
    netPaymentSet = MoneyBag
    netPaymentSet_presentmentMoney_amount = DeepExtract(["netPaymentSet", "presentmentMoney", "amount"], "DECIMAL")
    netPaymentSet_presentmentMoney_currencyCode = DeepExtract(
        ["netPaymentSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    netPaymentSet_shopMoney_amount = DeepExtract(["netPaymentSet", "shopMoney", "amount"], "DECIMAL")
    netPaymentSet_shopMoney_currencyCode = DeepExtract(["netPaymentSet", "shopMoney", "currencyCode"], "TEXT")
    # nonFulfillableLineItems = "nonFulfillableLineItems"
    note = "note"
    number = "number"
    originalTotalAdditionalFeesSet = MoneyBag
    originalTotalAdditionalFeesSet_presentmentMoney_amount = DeepExtract(
        ["originalTotalAdditionalFeesSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    originalTotalAdditionalFeesSet_presentmentMoney_currencyCode = DeepExtract(
        ["originalTotalAdditionalFeesSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    originalTotalAdditionalFeesSet_shopMoney_amount = DeepExtract(
        ["originalTotalAdditionalFeesSet", "shopMoney", "amount"], "DECIMAL"
    )
    originalTotalAdditionalFeesSet_shopMoney_currencyCode = DeepExtract(
        ["originalTotalAdditionalFeesSet", "shopMoney", "currencyCode"], "TEXT"
    )
    originalTotalDutiesSet = MoneyBag
    originalTotalDutiesSet_presentmentMoney_amount = DeepExtract(
        ["originalTotalDutiesSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    originalTotalDutiesSet_presentmentMoney_currencyCode = DeepExtract(
        ["originalTotalDutiesSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    originalTotalDutiesSet_shopMoney_amount = DeepExtract(["originalTotalDutiesSet", "shopMoney", "amount"], "DECIMAL")
    originalTotalDutiesSet_shopMoney_currencyCode = DeepExtract(
        ["originalTotalDutiesSet", "shopMoney", "currencyCode"], "TEXT"
    )
    originalTotalPriceSet = MoneyBag
    originalTotalPriceSet_presentmentMoney_amount = DeepExtract(
        ["originalTotalPriceSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    originalTotalPriceSet_presentmentMoney_currencyCode = DeepExtract(
        ["originalTotalPriceSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    originalTotalPriceSet_shopMoney_amount = DeepExtract(["originalTotalPriceSet", "shopMoney", "amount"], "DECIMAL")
    originalTotalPriceSet_shopMoney_currencyCode = DeepExtract(
        ["originalTotalPriceSet", "shopMoney", "currencyCode"], "TEXT"
    )
    # paymentCollectionDetails = "paymentCollectionDetails"
    paymentGatewayNames = "paymentGatewayNames"
    # paymentTerms = "paymentTerms"
    phone = "phone"
    poNumber = "poNumber"
    presentmentCurrencyCode = "presentmentCurrencyCode"
    processedAt = "processedAt"
    productNetwork = "productNetwork"
    # publication = "publication"
    # purchasingEntity = "purchasingEntity"
    refundable = "refundable"
    refundDiscrepancySet = MoneyBag
    refundDiscrepancySet_presentmentMoney_amount = DeepExtract(
        ["refundDiscrepancySet", "presentmentMoney", "amount"], "DECIMAL"
    )
    refundDiscrepancySet_presentmentMoney_currencyCode = DeepExtract(
        ["refundDiscrepancySet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    refundDiscrepancySet_shopMoney_amount = DeepExtract(["refundDiscrepancySet", "shopMoney", "amount"], "DECIMAL")
    refundDiscrepancySet_shopMoney_currencyCode = DeepExtract(
        ["refundDiscrepancySet", "shopMoney", "currencyCode"], "TEXT"
    )
    # refunds = "refunds"
    registeredSourceUrl = "registeredSourceUrl"
    requiresShipping = "requiresShipping"
    restockable = "restockable"
    # retailLocation = "retailLocation"
    # returns = "returns"
    returnStatus = "returnStatus"
    # risk = "risk"
    shippingAddress = MailingAddress
    # shippingLine = "shippingLine"
    # shippingLines = "shippingLines"
    # shopifyProtect = "shopifyProtect"
    sourceIdentifier = "sourceIdentifier"
    sourceName = "sourceName"
    # staffMember = "staffMember"
    statusPageUrl = "statusPageUrl"
    subtotalLineItemsQuantity = "subtotalLineItemsQuantity"
    subtotalPriceSet = MoneyBag
    subtotalPriceSet_presentmentMoney_amount = DeepExtract(
        ["subtotalPriceSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    subtotalPriceSet_presentmentMoney_currencyCode = DeepExtract(
        ["subtotalPriceSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    subtotalPriceSet_shopMoney_amount = DeepExtract(["subtotalPriceSet", "shopMoney", "amount"], "DECIMAL")
    subtotalPriceSet_shopMoney_currencyCode = DeepExtract(["subtotalPriceSet", "shopMoney", "currencyCode"], "TEXT")
    # suggestedRefund = "suggestedRefund"
    tags = "tags"
    taxesIncluded = "taxesIncluded"
    taxExempt = "taxExempt"
    # taxLines = "taxLines"
    test = "test"
    totalCapturableSet = MoneyBag
    totalCapturableSet_presentmentMoney_amount = DeepExtract(
        ["totalCapturableSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    totalCapturableSet_presentmentMoney_currencyCode = DeepExtract(
        ["totalCapturableSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    totalCapturableSet_shopMoney_amount = DeepExtract(["totalCapturableSet", "shopMoney", "amount"], "DECIMAL")
    totalCapturableSet_shopMoney_currencyCode = DeepExtract(["totalCapturableSet", "shopMoney", "currencyCode"], "TEXT")
    # totalCashRoundingAdjustment = "totalCashRoundingAdjustment"
    totalDiscountsSet = MoneyBag
    totalDiscountsSet_presentmentMoney_amount = DeepExtract(
        ["totalDiscountsSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    totalDiscountsSet_presentmentMoney_currencyCode = DeepExtract(
        ["totalDiscountsSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    totalDiscountsSet_shopMoney_amount = DeepExtract(["totalDiscountsSet", "shopMoney", "amount"], "DECIMAL")
    totalDiscountsSet_shopMoney_currencyCode = DeepExtract(["totalDiscountsSet", "shopMoney", "currencyCode"], "TEXT")
    totalOutstandingSet = MoneyBag
    totalOutstandingSet_presentmentMoney_amount = DeepExtract(
        ["totalOutstandingSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    totalOutstandingSet_presentmentMoney_currencyCode = DeepExtract(
        ["totalOutstandingSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    totalOutstandingSet_shopMoney_amount = DeepExtract(["totalOutstandingSet", "shopMoney", "amount"], "DECIMAL")
    totalOutstandingSet_shopMoney_currencyCode = DeepExtract(
        ["totalOutstandingSet", "shopMoney", "currencyCode"], "TEXT"
    )
    totalPriceSet = MoneyBag
    totalPriceSet_presentmentMoney_amount = DeepExtract(["totalPriceSet", "presentmentMoney", "amount"], "DECIMAL")
    totalPriceSet_presentmentMoney_currencyCode = DeepExtract(
        ["totalPriceSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    totalPriceSet_shopMoney_amount = DeepExtract(["totalPriceSet", "shopMoney", "amount"], "DECIMAL")
    totalPriceSet_shopMoney_currencyCode = DeepExtract(["totalPriceSet", "shopMoney", "currencyCode"], "TEXT")
    totalReceivedSet = MoneyBag
    totalReceivedSet_presentmentMoney_amount = DeepExtract(
        ["totalReceivedSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    totalReceivedSet_presentmentMoney_currencyCode = DeepExtract(
        ["totalReceivedSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    totalReceivedSet_shopMoney_amount = DeepExtract(["totalReceivedSet", "shopMoney", "amount"], "DECIMAL")
    totalReceivedSet_shopMoney_currencyCode = DeepExtract(["totalReceivedSet", "shopMoney", "currencyCode"], "TEXT")
    totalRefundedSet = MoneyBag
    totalRefundedSet_presentmentMoney_amount = DeepExtract(
        ["totalRefundedSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    totalRefundedSet_presentmentMoney_currencyCode = DeepExtract(
        ["totalRefundedSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    totalRefundedSet_shopMoney_amount = DeepExtract(["totalRefundedSet", "shopMoney", "amount"], "DECIMAL")
    totalRefundedSet_shopMoney_currencyCode = DeepExtract(["totalRefundedSet", "shopMoney", "currencyCode"], "TEXT")
    totalRefundedShippingSet = MoneyBag
    totalRefundedShippingSet_presentmentMoney_amount = DeepExtract(
        ["totalRefundedShippingSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    totalRefundedShippingSet_presentmentMoney_currencyCode = DeepExtract(
        ["totalRefundedShippingSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    totalRefundedShippingSet_shopMoney_amount = DeepExtract(
        ["totalRefundedShippingSet", "shopMoney", "amount"], "DECIMAL"
    )
    totalRefundedShippingSet_shopMoney_currencyCode = DeepExtract(
        ["totalRefundedShippingSet", "shopMoney", "currencyCode"], "TEXT"
    )
    totalShippingPriceSet = MoneyBag
    totalShippingPriceSet_presentmentMoney_amount = DeepExtract(
        ["totalShippingPriceSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    totalShippingPriceSet_presentmentMoney_currencyCode = DeepExtract(
        ["totalShippingPriceSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    totalShippingPriceSet_shopMoney_amount = DeepExtract(["totalShippingPriceSet", "shopMoney", "amount"], "DECIMAL")
    totalShippingPriceSet_shopMoney_currencyCode = DeepExtract(
        ["totalShippingPriceSet", "shopMoney", "currencyCode"], "TEXT"
    )
    totalTaxSet = MoneyBag
    totalTaxSet_presentmentMoney_amount = DeepExtract(["totalTaxSet", "presentmentMoney", "amount"], "DECIMAL")
    totalTaxSet_presentmentMoney_currencyCode = DeepExtract(["totalTaxSet", "presentmentMoney", "currencyCode"], "TEXT")
    totalTaxSet_shopMoney_amount = DeepExtract(["totalTaxSet", "shopMoney", "amount"], "DECIMAL")
    totalTaxSet_shopMoney_currencyCode = DeepExtract(["totalTaxSet", "shopMoney", "currencyCode"], "TEXT")
    totalTipReceivedSet = MoneyBag
    totalTipReceivedSet_presentmentMoney_amount = DeepExtract(
        ["totalTipReceivedSet", "presentmentMoney", "amount"], "DECIMAL"
    )
    totalTipReceivedSet_presentmentMoney_currencyCode = DeepExtract(
        ["totalTipReceivedSet", "presentmentMoney", "currencyCode"], "TEXT"
    )
    totalTipReceivedSet_shopMoney_amount = DeepExtract(["totalTipReceivedSet", "shopMoney", "amount"], "DECIMAL")
    totalTipReceivedSet_shopMoney_currencyCode = DeepExtract(
        ["totalTipReceivedSet", "shopMoney", "currencyCode"], "TEXT"
    )
    totalWeight = "totalWeight"
    # transactions = "transactions"
    transactionsCount = Count
    unpaid = "unpaid"
    updatedAt = "updatedAt"


columns = [
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "additionalFees",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of additional fees applied to an order, such as duties, import fees, or tax lines.",
    #     "IS_NULLABLE": False
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "agreements",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of sales agreements associated with the order, such as contracts defining payment terms, or delivery schedules between merchants and customers.",
    #     "IS_NULLABLE": False
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "alerts",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of messages that appear on the Orders page in the Shopify admin. These alerts provide merchants with important information about an order's status or required actions.",
    #     "IS_NULLABLE": False
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "app",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The application that created the order. For example, Online Store, Point of Sale, or a custom app name. Use this to identify the order source for attribution and fulfillment workflows.",
    #     "IS_NULLABLE": None
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "billingAddress",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The billing address associated with the payment method selected by the customer for an order. Returns null if no billing address was provided during checkout.",
    #     "IS_NULLABLE": None
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "billingAddressMatchesShippingAddress",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether the billing address matches the shipping address. Returns true if both addresses are the same, and false if they're different or if an address is missing.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "cancellation",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "Details of an order's cancellation, if it has been canceled. This includes the reason, date, and any staff notes.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "cancelledAt",
        "DATA_TYPE": "TIMESTAMP",
        "COLUMN_DESCRIPTION": "The date and time in ISO 8601 format when an order was canceled. Returns null if the order hasn't been canceled.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "cancelReason",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The reason provided for an order cancellation. For example, a merchant might cancel an order if there's insufficient inventory. Returns null if the order hasn't been canceled.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "canMarkAsPaid",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether an order can be manually marked as paid. Returns false if the order is already paid, is canceled, has pending Shopify Payments transactions, or has a negative payment amount.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "canNotifyCustomer",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether order notifications can be sent to the customer. Returns true if the customer has a valid email address.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "capturable",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether an authorized payment for an order can be captured. Returns true if an authorized payment exists that hasn't been fully captured yet.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "cartDiscountAmountSet",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The total discount amount applied at the time the order was created, displayed in both shop and presentment currencies, before returns, refunds, order edits, and cancellations. This field only includes discounts applied to the entire order.",
    #     "IS_NULLABLE": None
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "channelInformation",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "Details about the sales channel that created the order, such as the channel app type and channel name, which helps to track order sources.",
    #     "IS_NULLABLE": None
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "clientIp",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The IP address of the customer who placed the order. Useful for fraud detection and geographic analysis.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "closed",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether an order is closed. An order is considered closed if all its line items have been fulfilled or canceled, and all financial transactions are complete.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "closedAt",
        "DATA_TYPE": "TIMESTAMP",
        "COLUMN_DESCRIPTION": "The date and time ISO 8601 format when an order was closed. Shopify automatically records this timestamp when all items have been fulfilled or canceled, and all financial transactions are complete. Returns null if the order isn't closed.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "confirmationNumber",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "A customer-facing order identifier, often shown instead of the sequential order name. It uses a random alphanumeric format (for example, XPAV284CT) and isn't guaranteed to be unique across orders.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "confirmed",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether inventory has been reserved for an order. Returns true if inventory quantities for an order's line items have been reserved.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "createdAt",
        "DATA_TYPE": "TIMESTAMP",
        "COLUMN_DESCRIPTION": "The date and time in ISO 8601 format when an order was created. This timestamp is set when the customer completes checkout and remains unchanged throughout an order's lifecycle.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "currencyCode",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The shop currency when the order was placed. For example, USD or CAD.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "currentCartDiscountAmountSet",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The current total of all discounts applied to the entire order, after returns, refunds, order edits, and cancellations. This includes discount codes, automatic discounts, and other promotions that affect the whole order rather than individual line items.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "currentShippingPriceSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The current shipping price after applying refunds and discounts. If the parent order.taxesIncluded field is true, then this price includes taxes. Otherwise, this field is the pre-tax price.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "currentSubtotalLineItemsQuantity",
        "DATA_TYPE": "INT",
        "COLUMN_DESCRIPTION": "The current sum of the quantities for all line items that contribute to the order's subtotal price, after returns, refunds, order edits, and cancellations.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "currentSubtotalPriceSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total price of the order, after returns and refunds, in shop and presentment currencies. This includes taxes and discounts.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "currentTaxLines",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of all tax lines applied to line items on the order, after returns. Tax line prices represent the total price for all tax lines with the same rate and title.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "currentTotalAdditionalFeesSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The current total of all additional fees for an order, after any returns or modifications. Modifications include returns, refunds, order edits, and cancellations. Additional fees can include charges such as duties, import fees, and special handling.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "currentTotalDiscountsSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total amount discounted on the order after returns and refunds, in shop and presentment currencies. This includes both order and line level discounts.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "currentTotalDutiesSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The current total duties amount for an order, after any returns or modifications. Modifications include returns, refunds, order edits, and cancellations.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "currentTotalPriceSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total price of the order, after returns, in shop and presentment currencies. This includes taxes and discounts.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "currentTotalTaxSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The sum of the prices of all tax lines applied to line items on the order, after returns and refunds, in shop and presentment currencies.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "currentTotalWeight",
        "DATA_TYPE": "INT",
        "COLUMN_DESCRIPTION": "The total weight of the order after returns and refunds, in grams.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "customAttributes",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of additional information that has been attached to the order. For example, gift message, delivery instructions, or internal notes.",
    #     "IS_NULLABLE": False
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "customer",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The customer who placed an order. Returns null if an order was created through a checkout without customer authentication, such as a guest checkout.",
    #     "IS_NULLABLE": None
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "customerId",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "ID of the customer.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "customerAcceptsMarketing",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether the customer agreed to receive marketing emails at the time of purchase. Use this to ensure compliance with marketing consent laws and to segment customers for email campaigns.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "customerJourneySummary",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The customer's visits and interactions with the online store before placing the order. Use this to understand customer behavior, attribution sources, and marketing effectiveness to optimize your sales funnel.",
    #     "IS_NULLABLE": None
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "customerLocale",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The customer's language and region preference at the time of purchase. For example, en for English, fr-CA for French (Canada), or es-MX for Spanish (Mexico). Use this to provide localized customer service and targeted marketing in the customer's preferred language.",
        "IS_NULLABLE": None,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "discountApplications",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of discounts that are applied to the order, excluding order edits and refunds. Includes discount codes, automatic discounts, and other promotions that reduce the order total.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "discountCode",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The discount code used for an order. Returns null if no discount code was applied.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "discountCodes",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The discount codes used for the order. Multiple codes can be applied to a single order.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "displayAddress",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The primary address of the customer, prioritizing shipping address over billing address when both are available. Returns null if neither shipping address nor billing address was provided.",
    #     "IS_NULLABLE": None
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "displayFinancialStatus",
    #     "DATA_TYPE": "TEXT",
    #     "COLUMN_DESCRIPTION": "An order's financial status for display in the Shopify admin.",
    #     "IS_NULLABLE": None
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "displayFulfillmentStatus",
    #     "DATA_TYPE": "TEXT",
    #     "COLUMN_DESCRIPTION": "The order's fulfillment status that displays in the Shopify admin to merchants. For example, an order might be unfulfilled or scheduled.",
    #     "IS_NULLABLE": False
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "disputes",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of payment disputes associated with the order, such as chargebacks or payment inquiries. Disputes occur when customers challenge transactions with their bank or payment provider.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "dutiesIncluded",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether duties are included in the subtotal price of the order. Duties are import taxes charged by customs authorities when goods cross international borders.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "edited",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether the order has had any edits applied. For example, adding or removing line items, updating quantities, or changing prices.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "email",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The email address associated with the customer for this order. Used for sending order confirmations, shipping notifications, and other order-related communications. Returns null if no email address was provided during checkout.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "estimatedTaxes",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether taxes on the order are estimated. This field returns false when taxes on the order are finalized and aren't subject to any changes.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "events",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of events associated with the order. Events track significant changes and activities related to the order, such as creation, payment, fulfillment, and cancellation.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "fulfillable",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether there are line items that can be fulfilled. This field returns false when the order has no fulfillable line items.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "fulfillmentOrders",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of fulfillment orders for an order. Each fulfillment order groups line items that are fulfilled together, allowing an order to be processed in parts if needed.",
    #     "IS_NULLABLE": False
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "fulfillments",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of shipments for the order. Fulfillments represent the physical shipment of products to customers.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "fulfillmentsCount",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total number of fulfillments for the order, including canceled ones.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "fullyPaid",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether the order has been paid in full. This field returns true when the total amount received equals or exceeds the order total.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "hasTimelineComment",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether the merchant has added a timeline comment to the order.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "id",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "A globally-unique ID.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "legacyResourceId",
    #     "DATA_TYPE": "INT",
    #     "COLUMN_DESCRIPTION": "The ID of the corresponding resource in the REST Admin API.",
    #     "IS_NULLABLE": False
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "lineItems",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of the order's line items. Line items represent the individual products and quantities that make up the order.",
    #     "IS_NULLABLE": False
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "localizedFields",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "List of localized fields for the resource.",
    #     "IS_NULLABLE": False
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "merchantBusinessEntity",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The legal business structure that the merchant operates under for this order, such as an LLC, corporation, or partnership. Used for tax reporting, legal compliance, and determining which business entity is responsible for the order.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "merchantEditable",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether the order can be edited by the merchant. Returns false for orders that can't be modified, such as canceled orders or orders with specific payment statuses.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "merchantEditableErrors",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "A list of reasons why the order can't be edited. For example, canceled orders can't be edited.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "merchantOfRecordApp",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The application acting as the Merchant of Record for the order. The Merchant of Record is responsible for tax collection and remittance.",
    #     "IS_NULLABLE": None
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "metafield",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A custom field, including its namespace and key, that's associated with a Shopify resource for the purposes of adding and storing additional information.",
    #     "IS_NULLABLE": None
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "metafields",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of custom fields that a merchant associates with a Shopify resource.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "name",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The unique identifier for the order that appears on the order page in the Shopify admin and the Order status page. For example, #1001, EN1001, or 1001-A. This value isn't unique across multiple stores.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "netPaymentSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The net payment for the order, based on the total amount received minus the total amount refunded, in shop and presentment currencies.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "nonFulfillableLineItems",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of line items that can't be fulfilled. For example, tips and fully refunded line items can't be fulfilled.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "note",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The note associated with the order. Contains additional information or instructions added by merchants or customers during the order process. Commonly used for special delivery instructions, gift messages, or internal processing notes.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "number",
        "DATA_TYPE": "INT",
        "COLUMN_DESCRIPTION": "The order number used to generate the name using the store's configured order number prefix/suffix. This number isn't guaranteed to follow a consecutive integer sequence, nor is it guaranteed to be unique across multiple stores, or even for a single store.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "originalTotalAdditionalFeesSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total amount of all additional fees, such as import fees or taxes, that were applied when an order was created. Returns null if additional fees aren't applicable.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "originalTotalDutiesSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total amount of duties calculated when an order was created, before any modifications. Modifications include returns, refunds, order edits, and cancellations.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "originalTotalPriceSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total price of the order at the time of order creation, in shop and presentment currencies. Use this to compare the original order value against the current total after edits, returns, or refunds.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "paymentCollectionDetails",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The payment collection details for the order, including payment status, outstanding amounts, and collection information. Use this to understand when and how payments should be collected, especially for orders with deferred or installment payment terms.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "paymentGatewayNames",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "A list of the names of all payment gateways used for the order. For example, Shopify Payments and Cash on Delivery (COD).",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "paymentTerms",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The payment terms associated with the order, such as net payment due dates or early payment discounts. Payment terms define when and how an order should be paid. Returns null if no specific payment terms were set for the order.",
    #     "IS_NULLABLE": None
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "phone",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The phone number associated with the customer for this order. Useful for contacting customers about shipping updates, delivery notifications, or order issues. Returns null if no phone number was provided during checkout.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "poNumber",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The purchase order (PO) number that's associated with an order. This is typically provided by business customers who require a PO number for their procurement.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "presentmentCurrencyCode",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The currency used by the customer when placing the order. For example, USD, EUR, or CAD. This may differ from the shop's base currency when serving international customers or using multi-currency pricing.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "processedAt",
        "DATA_TYPE": "TIMESTAMP",
        "COLUMN_DESCRIPTION": "The date and time in ISO 8601 format when the order was processed. This date and time might not match the date and time when the order was created.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "productNetwork",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether the customer also purchased items from other stores in the network.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "publication",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The sales channel that the order was created from, such as the Online Store or Shopify POS.",
    #     "IS_NULLABLE": None
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "purchasingEntity",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The business entity that placed the order, including company details and purchasing relationships. Used for B2B transactions to track which company or organization is responsible for the purchase and payment terms.",
    #     "IS_NULLABLE": None
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "refundable",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether the order can be refunded based on its payment transactions. Returns false for orders with no eligible payment transactions, such as fully refunded orders or orders with non-refundable payment methods.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "refundDiscrepancySet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The difference between the suggested and actual refund amount of all refunds that have been applied to the order. A positive value indicates a difference in the merchant's favor, and a negative value indicates a difference in the customer's favor.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "refunds",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of refunds that have been applied to the order. Refunds represent money returned to customers for returned items, cancellations, or adjustments.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "registeredSourceUrl",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The URL of the source that the order originated from, if found in the domain registry. Returns null if the source URL isn't in the domain registry.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "requiresShipping",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether the order requires physical shipping to the customer. Returns false for digital-only orders (such as gift cards or downloadable products) and true for orders with physical products that need delivery.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "restockable",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether any line items on the order can be restocked into inventory. Returns false for digital products, custom items, or items that can't be resold.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "retailLocation",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The physical location where a retail order is created or completed, except for draft POS orders completed using the mark as paid flow in the Shopify admin, which return null. Transactions associated with the order might have been processed at a different location.",
    #     "IS_NULLABLE": None
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "returns",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The returns associated with the order. Contains information about items that customers have requested to return, including return reasons, status, and refund details.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "returnStatus",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The order's aggregated return status for display purposes. Indicates the overall state of returns for the order, helping merchants track and manage the return process.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "risk",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The risk assessment summary for the order. Provides fraud analysis and risk scoring to help you identify potentially fraudulent orders.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "shippingAddress",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The shipping address where the order will be delivered. Contains the customer's delivery location for fulfillment and shipping label generation. Returns null for digital orders or orders that don't require shipping.",
        "IS_NULLABLE": None,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "shippingLine",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A summary of all shipping costs on the order. Aggregates shipping charges, discounts, and taxes to provide a single view of delivery costs.",
    #     "IS_NULLABLE": None
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "shippingLines",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The shipping methods applied to the order. Each shipping line represents a shipping option chosen during checkout, including the carrier, service level, and cost.",
    #     "IS_NULLABLE": False
    # },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "shopifyProtect",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The Shopify Protect details for the order, including fraud protection status and coverage information. Shopify Protect helps protect eligible orders against fraudulent chargebacks. Returns null if Shopify Protect is disabled for the shop or the order isn't eligible for protection.",
    #     "IS_NULLABLE": None
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "sourceIdentifier",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "A unique POS or third party order identifier. For example, 1234-12-1000 or 111-98567-54.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "sourceName",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The name of the source associated with the order, such as web, mobile_app, or pos. Use this field to identify the platform where the order was placed.",
        "IS_NULLABLE": None,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "staffMember",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The staff member who created or is responsible for the order. Useful for tracking which team member handled phone orders, manual orders, or order modifications. Returns null for orders created directly by customers through the online store.",
    #     "IS_NULLABLE": None
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "statusPageUrl",
        "DATA_TYPE": "TEXT",
        "COLUMN_DESCRIPTION": "The URL where customers can check their order's current status, including tracking information and delivery updates.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "subtotalLineItemsQuantity",
        "DATA_TYPE": "INT",
        "COLUMN_DESCRIPTION": "The sum of quantities for all line items that contribute to the order's subtotal price. This excludes quantities for items like tips, shipping costs, or gift cards that don't affect the subtotal.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "subtotalPriceSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The sum of the prices for all line items after discounts and before returns, in shop and presentment currencies. If taxesIncluded is true, then the subtotal also includes tax.",
        "IS_NULLABLE": None,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "suggestedRefund",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A calculated refund suggestion for the order based on specified line items, shipping, and duties. Use this to preview refund amounts, taxes, and processing fees before creating an actual refund.",
    #     "IS_NULLABLE": None
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "tags",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "A comma separated list of tags associated with the order. Updating tags overwrites any existing tags that were previously added to the order.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "taxesIncluded",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether taxes are included in the subtotal price of the order. When true, the subtotal and line item prices include tax amounts. When false, taxes are calculated and displayed separately.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "taxExempt",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether taxes are exempt on the order. Returns true for orders where the customer or business has a valid tax exemption, such as non-profit organizations or tax-free purchases.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "taxLines",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of all tax lines applied to line items on the order, before returns. Tax line prices represent the total price for all tax lines with the same rate and title.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "test",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether the order is a test. Test orders are made using the Shopify Bogus Gateway or a payment provider with test mode enabled. A test order can't be converted into a real order and vice versa.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "totalCapturableSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The authorized amount that's uncaptured or undercaptured, in shop and presentment currencies. This amount isn't adjusted for returns.",
        "IS_NULLABLE": False,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "totalCashRoundingAdjustment",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "The total rounding adjustment applied to payments or refunds for an order involving cash payments. Applies to some countries where cash transactions are rounded to the nearest currency denomination.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "totalDiscountsSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total amount discounted on the order before returns, in shop and presentment currencies. This includes both order and line level discounts.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "totalOutstandingSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total amount not yet transacted for the order, in shop and presentment currencies. A positive value indicates a difference in the merchant's favor (payment from customer to merchant) and a negative value indicates a difference in the customer's favor (refund from merchant to customer).",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "totalPriceSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total price of the order, before returns, in shop and presentment currencies. This includes taxes and discounts.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "totalReceivedSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total amount received from the customer before returns, in shop and presentment currencies.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "totalRefundedSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total amount that was refunded, in shop and presentment currencies.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "totalRefundedShippingSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total amount of shipping that was refunded, in shop and presentment currencies.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "totalShippingPriceSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total shipping costs returned to the customer, in shop and presentment currencies. This includes fees and any related discounts that were refunded.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "totalTaxSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The total tax amount before returns, in shop and presentment currencies.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "totalTipReceivedSet",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The sum of all tip amounts for the order, in shop and presentment currencies.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "totalWeight",
        "DATA_TYPE": "INT",
        "COLUMN_DESCRIPTION": "The total weight of the order before returns, in grams.",
        "IS_NULLABLE": None,
    },
    # {
    #     "TABLE_NAME": "orders",
    #     "COLUMN_NAME": "transactions",
    #     "DATA_TYPE": "JSON",
    #     "COLUMN_DESCRIPTION": "A list of transactions associated with the order.",
    #     "IS_NULLABLE": False
    # },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "transactionsCount",
        "DATA_TYPE": "JSON",
        "COLUMN_DESCRIPTION": "The number of transactions associated with the order.",
        "IS_NULLABLE": None,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "unpaid",
        "DATA_TYPE": "BOOLEAN",
        "COLUMN_DESCRIPTION": "Whether no payments have been made for the order.",
        "IS_NULLABLE": False,
    },
    {
        "TABLE_NAME": "orders",
        "COLUMN_NAME": "updatedAt",
        "DATA_TYPE": "TIMESTAMP",
        "COLUMN_DESCRIPTION": "The date and time in ISO 8601 format when the order was last modified.",
        "IS_NULLABLE": False,
    },
]

# region add fields flattened from JSON
for field_name, field in Orders.aliases():
    if isinstance(field, DeepExtract):
        columns.append(
            {
                "TABLE_NAME": "orders",
                "COLUMN_NAME": field_name,
                "DATA_TYPE": field.mysql_data_type,
                "COLUMN_DESCRIPTION": field.description,
                "IS_NULLABLE": None,
            }
        )
# endregion



import os
import json
import hashlib
from collections import OrderedDict
from typing import Dict, List, Literal, Tuple
from urllib.parse import urlparse

import pandas as pd
import psycopg
from psycopg import sql
from mindsdb_sql_parser.ast import (
    Parameter,
    Identifier,
    BinaryOperation,
    Tuple as AstTuple,
    Constant,
    Select,
    OrderBy,
    TypeCast,
    Delete,
    Update,
    Function,
    DropTables,
)
from mindsdb_sql_parser.ast.base import ASTNode
from pgvector.psycopg import register_vector

from mindsdb.integrations.handlers.postgres_handler.postgres_handler import (
    PostgresHandler,
)
from mindsdb.integrations.libs.response import RESPONSE_TYPE, HandlerResponse as Response
from mindsdb.integrations.libs.vectordatabase_handler import (
    FilterCondition,
    VectorStoreHandler,
    DistanceFunction,
    TableField,
    FilterOperator,
)
from mindsdb.integrations.libs.keyword_search_base import KeywordSearchBase
from mindsdb.integrations.utilities.sql_utils import KeywordSearchArgs
from mindsdb.utilities import log
from mindsdb.utilities.profiler import profiler
from mindsdb.utilities.context import context as ctx

logger = log.getLogger(__name__)


# todo Issue #7316 add support for different indexes and search algorithms e.g. cosine similarity or L2 norm
class PgVectorHandler(PostgresHandler, VectorStoreHandler, KeywordSearchBase):
    """This handler handles connection and execution of the PostgreSQL with pgvector extension statements."""

    name = "pgvector"

    # LRU cache to track which legacy tables have been checked for migration.
    # This prevents repeated database existence checks on every table access.
    _migration_checked_cache: OrderedDict = OrderedDict()
    _MIGRATION_CACHE_MAXSIZE = 1024

    def __init__(self, name: str, **kwargs):
        super().__init__(name=name, **kwargs)
        self._is_shared_db = False
        self._is_vector_registered = False
        # we get these from the connection args on PostgresHandler parent
        self._is_sparse = self.connection_args.get("is_sparse", False)
        self._vector_size = self.connection_args.get("vector_size", None)

        if self._is_sparse:
            if not self._vector_size:
                raise ValueError("vector_size is required when is_sparse=True")

                # Use inner product for sparse vectors
                distance_op = "<#>"

        else:
            distance_op = "<=>"
            if "distance" in self.connection_args:
                distance_ops = {
                    "l1": "<+>",
                    "l2": "<->",
                    "ip": "<#>",  # inner product
                    "cosine": "<=>",
                    "hamming": "<~>",
                    "jaccard": "<%>",
                }

                distance_op = distance_ops.get(self.connection_args["distance"])
                if distance_op is None:
                    raise ValueError(f"Wrong distance type. Allowed options are {list(distance_ops.keys())}")

        self.distance_op = distance_op
        self.connect()

    def get_metric_type(self) -> str:
        """
        Get the metric type from the distance ops

        """
        distance_ops_to_metric_type_map = {
            "<->": "vector_l2_ops",
            "<#>": "vector_ip_ops",
            "<=>": "vector_cosine_ops",
            "<+>": "vector_l1_ops",
            "<~>": "bit_hamming_ops",
            "<%>": "bit_jaccard_ops",
        }
        return distance_ops_to_metric_type_map.get(self.distance_op, "vector_cosine_ops")

    def _make_connection_args(self):
        cloud_pgvector_url = os.environ.get("KB_PGVECTOR_URL")
        # if no connection args and shared pg vector defined - use it
        if len(self.connection_args) == 0 and cloud_pgvector_url is not None:
            result = urlparse(cloud_pgvector_url)
            self.connection_args = {
                "host": result.hostname,
                "port": result.port,
                "user": result.username,
                "password": result.password,
                "database": result.path[1:],
            }
            self._is_shared_db = True
        return super()._make_connection_args()

    def get_tables(self) -> Response:
        # Hide list of tables from all users
        if self._is_shared_db:
            return Response(RESPONSE_TYPE.OK)
        return super().get_tables()

    def query(self, query: ASTNode) -> Response:
        # Option to drop table of shared pgvector connection
        if isinstance(query, DropTables):
            query.tables = [self._check_table(table.parts[-1]) for table in query.tables]
            query_str, params = self.renderer.get_exec_params(query, with_failback=True)
            return self.native_query(query_str, params, no_restrict=True)
        return super().query(query)

    def native_query(self, query, params=None, no_restrict=False) -> Response:
        """
        Altered `native_query` method of postgres handler.
        Restrict usage of native query from executor with shared pg vector connection
          Exceptions: if it is used by pgvector itself (with no_restrict = True)
        """
        # Prevent execute native queries
        if self._is_shared_db and not no_restrict:
            return Response(RESPONSE_TYPE.OK)
        return super().native_query(query, params=params)

    def raw_query(self, query, params=None) -> Response:
        resp = super().native_query(query, params)
        if resp.resp_type == RESPONSE_TYPE.ERROR:
            raise RuntimeError(resp.error_message)
        if resp.resp_type == RESPONSE_TYPE.TABLE:
            return resp.data_frame

    @profiler.profile()
    def connect(self) -> psycopg.connection:
        """
        Handles the connection to a PostgreSQL database instance.
        """
        self.connection = super().connect()
        if self._is_vector_registered:
            return self.connection

        with self.connection.cursor() as cur:
            try:
                # load pg_vector extension
                cur.execute("CREATE EXTENSION IF NOT EXISTS vector")
                logger.info("pg_vector extension loaded")

            except psycopg.Error as e:
                self.connection.rollback()
                logger.error(f"Error loading pg_vector extension, ensure you have installed it before running, {e}!")
                raise

        # register vector type with psycopg2 connection
        register_vector(self.connection)
        self._is_vector_registered = True

        return self.connection

    def add_full_text_index(self, table_name: str, column_name: str) -> Response:
        """
        Add a full text index to the specified column of the table.
        Args:
            table_name (str): Name of the table to add the index to.
            column_name (str): Name of the column to add the index to.
        Returns:
            Response: Response object indicating success or failure.
        """
        table_name = self._check_table(table_name)
        # Quote table name for PostgreSQL - table names may contain special chars like hyphens
        query = f'CREATE INDEX IF NOT EXISTS "{table_name}_{column_name}_fts_idx" ON "{table_name}" USING gin(to_tsvector(\'english\', {column_name}))'
        self.raw_query(query)
        return Response(RESPONSE_TYPE.OK)

    @staticmethod
    def _translate_conditions(conditions: List[FilterCondition]) -> Tuple[List[dict], dict]:
        """
        Translate filter conditions to a dictionary
        """

        if conditions is None:
            conditions = []

        filter_conditions = []
        embedding_condition = None

        for condition in conditions:
            is_embedding = condition.column == "embeddings"

            parts = condition.column.split(".")
            key = Identifier(parts[0])

            # converts 'col.el1.el2' to col->'el1'->>'el2'
            if len(parts) > 1:
                # intermediate elements
                for el in parts[1:-1]:
                    key = BinaryOperation(op="->", args=[key, Constant(el)])

                # last element
                key = BinaryOperation(op="->>", args=[key, Constant(parts[-1])])

            type_cast = None
            value = condition.value
            if (
                isinstance(value, list)
                and len(value) > 0
                and condition.op in (FilterOperator.IN, FilterOperator.NOT_IN)
            ):
                value = condition.value[0]

            if isinstance(value, int):
                type_cast = "int"
            elif isinstance(value, float):
                type_cast = "float"
            if type_cast is not None:
                key = TypeCast(type_cast, key)

            item = {
                "name": key,
                "op": condition.op.value,
                "value": condition.value,
            }
            if is_embedding:
                embedding_condition = item
            else:
                filter_conditions.append(item)

        return filter_conditions, embedding_condition

    @staticmethod
    def _construct_where_clause(filter_conditions=None):
        """
        Construct where clauses from filter conditions
        """

        where_clause = None

        for item in filter_conditions:
            key = item["name"]

            if item["op"].lower() in ("in", "not in"):
                values = [Constant(i) for i in item["value"]]
                value = AstTuple(values)
            else:
                value = Constant(item["value"])
            condition = BinaryOperation(op=item["op"], args=[key, value])

            if where_clause is None:
                where_clause = condition
            else:
                where_clause = BinaryOperation(op="AND", args=[where_clause, condition])
        return where_clause

    @staticmethod
    def _construct_full_after_from_clause(
        where_clause: str,
        offset_clause: str,
        limit_clause: str,
    ) -> str:
        return f"{where_clause} {offset_clause} {limit_clause}"

    def _build_keyword_bm25_query(
        self,
        table_name: str,
        keyword_search_args: KeywordSearchArgs,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        offset: int = None,
    ):
        if columns is None:
            columns = ["id", "content", "metadata"]

        filter_conditions, _ = self._translate_conditions(conditions)
        where_clause = self._construct_where_clause(filter_conditions)

        if keyword_search_args:
            keyword_query_condition = BinaryOperation(
                op="@@",
                args=[
                    Function("to_tsvector", args=[Constant("english"), Identifier(keyword_search_args.column)]),
                    Function("websearch_to_tsquery", args=[Constant("english"), Constant(keyword_search_args.query)]),
                ],
            )

            if where_clause:
                where_clause = BinaryOperation(op="AND", args=[where_clause, keyword_query_condition])
            else:
                where_clause = keyword_query_condition

        distance = Function(
            "ts_rank_cd",
            args=[
                Function("to_tsvector", args=[Constant("english"), Identifier(keyword_search_args.column)]),
                Function("websearch_to_tsquery", args=[Constant("english"), Constant(keyword_search_args.query)]),
            ],
            alias=Identifier("distance"),
        )

        targets = [Identifier(col) for col in columns]
        targets.append(distance)

        limit_clause = Constant(limit) if limit else None
        offset_clause = Constant(offset) if offset else None

        return Select(
            targets=targets,
            from_table=Identifier(table_name),
            where=where_clause,
            limit=limit_clause,
            offset=offset_clause,
        )

    def _build_select_query(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        offset: int = None,
    ) -> Select:
        """
        given inputs, build string query
        """
        limit_clause = Constant(limit) if limit else None
        offset_clause = Constant(offset) if offset else None

        # translate filter conditions to dictionary
        filter_conditions, embedding_search = self._translate_conditions(conditions)

        # given filter conditions, construct where clause
        where_clause = self._construct_where_clause(filter_conditions)

        # Handle distance column specially since it's calculated, not stored
        modified_columns = []
        has_distance = False
        if columns is not None:
            for col in columns:
                if col == TableField.DISTANCE.value:
                    has_distance = True
                else:
                    modified_columns.append(col)
        else:
            modified_columns = ["id", "content", "embeddings", "metadata"]
            has_distance = True

        targets = [Identifier(col) for col in modified_columns]

        query = Select(
            targets=targets,
            from_table=Identifier(table_name),
            where=where_clause,
            limit=limit_clause,
            offset=offset_clause,
        )

        if embedding_search:
            search_vector = embedding_search["value"]

            if self._is_sparse:
                # Convert dict to sparse vector if needed
                if isinstance(search_vector, dict):
                    from pgvector.utils import SparseVector

                    embedding = SparseVector(search_vector, self._vector_size)
                    search_vector = embedding.to_text()
            else:
                # Convert list to vector string if needed
                if isinstance(search_vector, list):
                    search_vector = f"[{','.join(str(x) for x in search_vector)}]"

            vector_op = BinaryOperation(
                op=self.distance_op,
                args=[Identifier("embeddings"), Constant(search_vector)],
                alias=Identifier("distance"),
            )
            # Calculate distance as part of the query if needed
            if has_distance:
                query.targets.append(vector_op)

            query.order_by = [OrderBy(vector_op, direction="ASC")]

        return query

    def _check_table(self, table_name: str):
        # Apply namespace for a user if the database is shared
        if self._is_shared_db:
            company_id = ctx.company_id
            user_id = ctx.user_id

            # PostgreSQL has a 63-character limit for identifiers.
            # Using full UUIDs (36 chars each) would exceed this limit.
            # We use a SHA-256 hash of company_id + user_id to create a shorter, unique prefix.
            namespace_hash = hashlib.sha256(f"{company_id}_{user_id}".encode()).hexdigest()[:16]
            new_table_name = f"t_{namespace_hash}_{table_name}"

            # Backwards compatibility: migrate old tables to the new hashed format
            # Old format: t_{company_id}_{table_name} (without user_id)
            old_table_name = f"t_{company_id}_{table_name}"
            self._migrate_legacy_table(old_table_name, new_table_name)

            return new_table_name
        return table_name

    def _migrate_legacy_table(self, old_name: str, new_name: str):
        """
        Rename legacy tables from t_{company_id}_{name} to t_{company_id}_{user_id}_{name}.

        Uses an LRU cache to ensure migration is only attempted once per table pair per process,
        avoiding repeated database existence checks on every table access.

        Args:
            old_name (str): The name of the old table.
            new_name (str): The name of the new table.

        Raises:
            RuntimeError: If the table rename fails.
        """
        cache_key = (old_name, new_name)

        # Check LRU cache - if already checked, return early
        if cache_key in PgVectorHandler._migration_checked_cache:
            # Move to end (mark as recently used)
            PgVectorHandler._migration_checked_cache.move_to_end(cache_key)
            return

        # Check if old table exists
        # Note: Use %s without quotes - psycopg handles proper escaping for string values
        old_exists_df = self.raw_query(
            "SELECT EXISTS (SELECT FROM information_schema.tables WHERE table_name = %s)", [[old_name]]
        )
        old_exists = old_exists_df.iloc[0, 0] if old_exists_df is not None and not old_exists_df.empty else False

        try:
            if old_exists:
                # Check if new table already exists
                new_exists_df = self.raw_query(
                    "SELECT EXISTS (SELECT FROM information_schema.tables WHERE table_name = %s)", [[new_name]]
                )
                new_exists = (
                    new_exists_df.iloc[0, 0] if new_exists_df is not None and not new_exists_df.empty else False
                )

                if not new_exists:
                    logger.info(f"Migrating legacy pgvector table {old_name} to {new_name}")
                    rename_query = sql.SQL("ALTER TABLE {} RENAME TO {}").format(
                        sql.Identifier(f'"{old_name}"'), sql.Identifier(f'"{new_name}"')
                    )
                    self.raw_query(rename_query.as_string(self.connection))
        except Exception:
            logger.exception(f"Failed to migrate legacy pgvector table {old_name} to {new_name}")
            raise

        # Add to LRU cache with eviction if at capacity
        if len(PgVectorHandler._migration_checked_cache) >= PgVectorHandler._MIGRATION_CACHE_MAXSIZE:
            PgVectorHandler._migration_checked_cache.popitem(last=False)  # Remove oldest
        PgVectorHandler._migration_checked_cache[cache_key] = True

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ) -> pd.DataFrame:
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """
        table_name = self._check_table(table_name)

        if columns is None:
            columns = ["id", "content", "embeddings", "metadata"]

        query = self._build_select_query(table_name, columns, conditions, limit, offset)
        query_str = self.renderer.get_string(query, with_failback=True)
        result = self.raw_query(query_str)

        # ensure embeddings are returned as string so they can be parsed by mindsdb
        if "embeddings" in columns:
            result["embeddings"] = result["embeddings"].apply(list)

        return result

    def keyword_select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
        keyword_search_args: KeywordSearchArgs = None,
    ) -> pd.DataFrame:
        table_name = self._check_table(table_name)

        if columns is None:
            columns = ["id", "content", "embeddings", "metadata"]

        query = self._build_keyword_bm25_query(table_name, keyword_search_args, columns, conditions, limit, offset)
        query_str = self.renderer.get_string(query, with_failback=True)
        result = self.raw_query(query_str)

        # ensure embeddings are returned as string so they can be parsed by mindsdb
        if "embeddings" in columns:
            result["embeddings"] = result["embeddings"].astype(str)

        return result

    def hybrid_search(
        self,
        table_name: str,
        embeddings: List[float],
        query: str = None,
        metadata: Dict[str, str] = None,
        distance_function=DistanceFunction.COSINE_DISTANCE,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Executes a hybrid search, combining semantic search and one or both of keyword/metadata search.

        For insight on the query construction, see: https://docs.pgvecto.rs/use-case/hybrid-search.html#advanced-search-merge-the-results-of-full-text-search-and-vector-search.

        Args:
            table_name(str): Name of underlying table containing content, embeddings, & metadata
            embeddings(List[float]): Embedding vector to perform semantic search against
            query(str): User query to convert into keywords for keyword search
            metadata(Dict[str, str]): Metadata filters to filter content rows against
            distance_function(DistanceFunction): Distance function used to compare embeddings vectors for semantic search

        Kwargs:
            id_column_name(str): Name of ID column in underlying table
            content_column_name(str): Name of column containing document content in underlying table
            embeddings_column_name(str): Name of column containing embeddings vectors in underlying table
            metadata_column_name(str): Name of column containing metadata key-value pairs in underlying table

        Returns:
            df(pd.DataFrame): Hybrid search result, sorted by hybrid search rank
        """
        if query is None and metadata is None:
            raise ValueError(
                "Must provide at least one of: query for keyword search, or metadata filters. For only embeddings search, use normal search instead."
            )

        id_column_name = kwargs.get("id_column_name", "id")
        content_column_name = kwargs.get("content_column_name", "content")
        embeddings_column_name = kwargs.get("embeddings_column_name", "embeddings")
        metadata_column_name = kwargs.get("metadata_column_name", "metadata")
        # Filter by given metadata for semantic search & full text search CTEs, if present.
        where_clause = " WHERE "
        if metadata is None:
            where_clause = ""
            metadata = {}
        for i, (k, v) in enumerate(metadata.items()):
            where_clause += f"{metadata_column_name}->>'{k}' = '{v}'"
            if i < len(metadata.items()) - 1:
                where_clause += " AND "

        # See https://docs.pgvecto.rs/use-case/hybrid-search.html#advanced-search-merge-the-results-of-full-text-search-and-vector-search.
        #
        # We can break down the below query as follows:
        #
        # Start with a CTE (Common Table Expression) called semantic_search (https://www.postgresql.org/docs/current/queries-with.html).
        # This expression calculates rank by the defined distance function, which measures the distance between the
        # embeddings column and the given embeddings vector. Results are ordered by this rank.
        #
        # Next, define another CTE called full_text_search if we are doing keyword search.
        # This calculates rank using the built-in ts_rank function (https://www.postgresql.org/docs/current/textsearch-controls.html#TEXTSEARCH-RANKING).
        # We convert the content column to a ts_vector and match rows for the given tsquery in the content column. Results are ordered by this ts_rank.
        #
        # For both of these CTEs, we filter by any given metadata fields.
        #
        # See https://www.postgresql.org/docs/current/textsearch-controls.html#TEXTSEARCH-PARSING-DOCUMENTS for to_tsvector
        # See https://www.postgresql.org/docs/current/functions-textsearch.html#FUNCTIONS-TEXTSEARCH for tsquery syntax
        #
        # Finally, we use a FULL OUTER JOIN to SELECT from both CTEs defined above.
        # The COALESCE function is used to handle cases where one CTE has null values.
        #
        # Or, if we are only doing metadata search, we leave out the JOIN & full text search CTEs.
        #
        # We calculate the final "hybrid" rank by summing the reciprocals of the ranks from each individual CTE.
        semantic_search_cte = f"""WITH semantic_search AS (
    SELECT {id_column_name}, {content_column_name}, {embeddings_column_name},
    RANK () OVER (ORDER BY {embeddings_column_name} {distance_function.value} '{str(embeddings)}') AS rank
    FROM "{table_name}" {where_clause}
    ORDER BY {embeddings_column_name} {distance_function.value} '{str(embeddings)}'::vector
    )"""

        full_text_search_cte = ""
        if query is not None:
            ts_vector_clause = (
                f"WHERE to_tsvector('english', {content_column_name}) @@ plainto_tsquery('english', '{query}')"
            )
            if metadata:
                ts_vector_clause = (
                    f"AND to_tsvector('english', {content_column_name}) @@ plainto_tsquery('english', '{query}')"
                )
            full_text_search_cte = f""",
    full_text_search AS (
    SELECT {id_column_name}, {content_column_name}, {embeddings_column_name},
    RANK () OVER (ORDER BY ts_rank(to_tsvector('english', {content_column_name}), plainto_tsquery('english', '{query}')) DESC) AS rank
    FROM "{table_name}" {where_clause}
    {ts_vector_clause}
    ORDER BY ts_rank(to_tsvector('english', {content_column_name}), plainto_tsquery('english', '{query}')) DESC
    )"""

        hybrid_select = """
    SELECT * FROM semantic_search"""
        if query is not None:
            hybrid_select = f"""
    SELECT
        COALESCE(semantic_search.{id_column_name}, full_text_search.{id_column_name}) AS id,
        COALESCE(semantic_search.{content_column_name}, full_text_search.{content_column_name}) AS content,
        COALESCE(semantic_search.{embeddings_column_name}, full_text_search.{embeddings_column_name}) AS embeddings,
        COALESCE(1.0 / (1 + semantic_search.rank), 0.0) + COALESCE(1.0 / (1 + full_text_search.rank), 0.0) AS rank
    FROM semantic_search FULL OUTER JOIN full_text_search USING ({id_column_name}) ORDER BY rank DESC;
        """

        full_search_query = f"{semantic_search_cte}{full_text_search_cte}{hybrid_select}"
        return self.raw_query(full_search_query)

    def create_table(self, table_name: str):
        """Create a table with a vector column."""

        table_name = self._check_table(table_name)

        with self.connection.cursor() as cur:
            # For sparse vectors, use sparsevec type
            vector_column_type = "sparsevec" if self._is_sparse else "vector"

            # Vector size is required for sparse vectors, optional for dense
            if self._is_sparse and not self._vector_size:
                raise ValueError("vector_size is required for sparse vectors")

            # Add vector size specification only if provided
            size_spec = f"({self._vector_size})" if self._vector_size is not None else "()"
            if vector_column_type == "vector":
                size_spec = ""

            cur.execute(f"""
                CREATE TABLE IF NOT EXISTS "{table_name}" (
                    id TEXT PRIMARY KEY,
                    embeddings {vector_column_type}{size_spec},
                    content TEXT,
                    metadata JSONB
                )
            """)
            self.connection.commit()

    def insert(self, table_name: str, data: pd.DataFrame):
        """
        Insert data into the pgvector table database.
        """
        table_name = self._check_table(table_name)

        if "metadata" in data.columns:
            data["metadata"] = data["metadata"].apply(json.dumps)

        resp = super().insert(table_name, data)
        if resp.resp_type == RESPONSE_TYPE.ERROR:
            raise RuntimeError(resp.error_message)
        if resp.resp_type == RESPONSE_TYPE.TABLE:
            return resp.data_frame

    def update(self, table_name: str, data: pd.DataFrame, key_columns: List[str] = None):
        """
        Udate data into the pgvector table database.
        """
        table_name = self._check_table(table_name)

        where = None
        update_columns = {}

        for col in data.columns:
            value = Parameter("%s")

            if col in key_columns:
                cond = BinaryOperation(op="=", args=[Identifier(col), value])
                if where is None:
                    where = cond
                else:
                    where = BinaryOperation(op="AND", args=[where, cond])
            else:
                update_columns[col] = value

        query = Update(table=Identifier(table_name), update_columns=update_columns, where=where)

        if TableField.METADATA.value in data.columns:

            def fnc(v):
                if isinstance(v, dict):
                    return json.dumps(v)

            data[TableField.METADATA.value] = data[TableField.METADATA.value].apply(fnc)

            data = data.astype({TableField.METADATA.value: str})

        transposed_data = []
        for _, record in data.iterrows():
            row = [record[col] for col in update_columns.keys()]
            for key_column in key_columns:
                row.append(record[key_column])
            transposed_data.append(row)

        query_str = self.renderer.get_string(query)
        self.raw_query(query_str, transposed_data)

    def delete(self, table_name: str, conditions: List[FilterCondition] = None):
        table_name = self._check_table(table_name)

        filter_conditions, _ = self._translate_conditions(conditions)
        where_clause = self._construct_where_clause(filter_conditions)

        query = Delete(table=Identifier(table_name), where=where_clause)
        query_str = self.renderer.get_string(query, with_failback=True)
        self.raw_query(query_str)

    def drop_table(self, table_name: str, if_exists=True):
        """
        Run a drop table query on the pgvector database.
        """
        table_name = self._check_table(table_name)
        self.raw_query(f'DROP TABLE IF EXISTS "{table_name}"')

    def create_index(
        self,
        table_name: str,
        column_name: str = "embeddings",
        index_type: Literal["ivfflat", "hnsw"] = "hnsw",
        metric_type: str = None,
    ):
        """
        Create an index on the pgvector table.
        Args:
            table_name (str): Name of the table to create the index on.
            column_name (str): Name of the column to create the index on.
            index_type (str): Type of the index to create. Supported types are 'ivfflat' and 'hnsw'.
            metric_type (str): Metric type for the index. Supported types are 'vector_l2_ops', 'vector_ip_ops', and 'vector_cosine_ops'.
        """
        if metric_type is None:
            metric_type = self.get_metric_type()
        # Check if the index type is supported
        if index_type not in ["ivfflat", "hnsw"]:
            raise ValueError("Invalid index type. Supported types are 'ivfflat' and 'hnsw'.")
        table_name = self._check_table(table_name)
        # first we make sure embedding dimension is set
        # Quote table name for PostgreSQL - table names may contain special chars like hyphens
        embedding_dim_size_df = self.raw_query(f'SELECT vector_dims({column_name}) FROM "{table_name}" LIMIT 1')
        # check if answer is empty
        if embedding_dim_size_df.empty:
            raise ValueError("Could not determine embedding dimension size. Make sure that knowledge base isn't empty")
        try:
            embedding_dim = int(embedding_dim_size_df.iloc[0, 0])
            # alter table to add dimension
            self.raw_query(f'ALTER TABLE "{table_name}" ALTER COLUMN {column_name} TYPE vector({embedding_dim})')
        except Exception:
            raise ValueError("Could not determine embedding dimension size. Make sure that knowledge base isn't empty")

        # Create the index
        self.raw_query(f'CREATE INDEX ON "{table_name}" USING {index_type} ({column_name} {metric_type})')



"""Custom LLM wrapper to replace langchain create_chat_model"""

from typing import List, Any, Dict
from dataclasses import dataclass

from mindsdb.interfaces.knowledge_base.llm_client import LLMClient
from mindsdb.utilities import log

logger = log.getLogger(__name__)


@dataclass
class LLMResponse:
    """Simple response object with content attribute to match langchain interface"""

    content: str


class CustomLLMWrapper:
    """
    Custom LLM wrapper that wraps LLMClient to provide langchain-compatible interface.
    This replaces langchain's create_chat_model for use in knowledge_base.
    """

    def __init__(self, args: Dict[str, Any], session=None):
        """
        Initialize the LLM wrapper

        Args:
            args: Dictionary with model_name, provider, and other LLM parameters
            session: Optional session for LLMClient
        """
        # Prepare params for LLMClient
        params = {
            "model_name": args.get("model_name"),
            "provider": args.get("provider", "openai"),
            **{k: v for k, v in args.items() if k not in ["model_name", "provider"]},
        }

        self.llm_client = LLMClient(params=params, session=session)

    def batch(self, prompts: List[str]) -> List[LLMResponse]:
        """
        Process a batch of prompts synchronously

        Args:
            prompts: List of prompt strings

        Returns:
            List of LLMResponse objects with content attribute
        """
        # Process each prompt separately since completion expects a single list of messages
        responses = []
        for prompt in prompts:
            # Convert prompt to messages format expected by LLMClient
            messages = [{"role": "user", "content": prompt}]

            # Call completion for this prompt's messages
            result = self.llm_client.completion(messages)

            # completion returns a list, get the first (and only) response
            if result:
                responses.append(LLMResponse(content=result[0]))
            else:
                responses.append(LLMResponse(content=""))

        return responses

    async def abatch(self, prompts: List[str]) -> List[LLMResponse]:
        """
        Process a batch of prompts asynchronously in parallel

        Args:
            prompts: List of prompt strings

        Returns:
            List of LLMResponse objects with content attribute
        """
        if not prompts:
            return []

        # Convert prompts to messages format expected by LLMClient
        messages_list = [[{"role": "user", "content": prompt}] for prompt in prompts]

        try:
            # Call abatch on LLMClient to process all prompts in parallel
            results = await self.llm_client.abatch(messages_list)

            # Convert results to LLMResponse objects
            responses = []
            for result in results:
                if result:
                    responses.append(LLMResponse(content=result[0]))
                else:
                    responses.append(LLMResponse(content=""))

            return responses
        except Exception as e:
            logger.error(f"Error processing prompts in abatch: {e}")
            # Return empty responses for all prompts on error
            return [LLMResponse(content="") for _ in prompts]


def create_chat_model(args: Dict[str, Any], session=None) -> CustomLLMWrapper:
    """
    Create a custom LLM wrapper (replacement for langchain's create_chat_model)

    Args:
        args: Dictionary with model_name, provider, and other LLM parameters
        session: Optional session for LLMClient

    Returns:
        CustomLLMWrapper instance
    """
    return CustomLLMWrapper(args, session=session)



import re
from typing import Union
from dataclasses import dataclass

from mindsdb.utilities import log

logger = log.getLogger(__name__)

# Default format instructions for conversational agent
# This is a simplified version - can be customized if needed
FORMAT_INSTRUCTIONS = """Use the following format:

Question: the input question you must answer
Thought: you should think about what to do
Action: the action to take, should be one of the available tools
Action Input: the input to the action
Observation: the result of the action
... (this Thought/Action/Action Input/Observation can repeat N times)
Thought: I now know the final answer
Final Answer: the final answer to the original input question"""


@dataclass
class AgentAction:
    """Custom AgentAction class to replace langchain AgentAction"""

    tool: str
    tool_input: str
    log: str


@dataclass
class AgentFinish:
    """Custom AgentFinish class to replace langchain AgentFinish"""

    return_values: dict
    log: str


class SafeOutputParser:
    """Output parser for the conversational agent that does not throw OutputParserException."""

    def __init__(self, ai_prefix: str = "AI", format_instructions: str = FORMAT_INSTRUCTIONS):
        self.ai_prefix = ai_prefix
        self.format_instructions = format_instructions

    def get_format_instructions(self) -> str:
        """Returns formatting instructions for the given output parser."""
        return self.format_instructions

    def parse(self, text: str) -> Union[AgentAction, AgentFinish]:
        """Parses outputted text from an LLM.

        Args:
            text (str): Outputted text to parse.

        Returns:
            Union[AgentAction, AgentFinish]: Parsed agent action or finish result
        """
        regex = r"Action: (.*?)[\n]*Action Input:([\s\S]*)"
        match = re.search(regex, text, re.DOTALL)
        if match is not None:
            action = match.group(1)
            action_input = match.group(2)
            return AgentAction(action.strip(), action_input.strip(" ").strip('"'), text)
        output = text
        if f"{self.ai_prefix}:" in text:
            output = text.split(f"{self.ai_prefix}:")[-1].strip()
        return AgentFinish({"output": output}, text)

    def extract_output(self, result: Union[AgentAction, AgentFinish, str]) -> str:
        """Extract the actual output text from a parse result.

        Args:
            result: Result from parse() method or a string

        Returns:
            str: The actual output text
        """
        if isinstance(result, str):
            return result
        elif isinstance(result, AgentFinish):
            return result.return_values.get("output", result.log)
        elif isinstance(result, AgentAction):
            # For AgentAction, return the log or tool_input
            return result.tool_input if result.tool_input else result.log
        else:
            return str(result)

    @property
    def _type(self) -> str:
        return "conversational"



import os
from textwrap import dedent
from typing import Any
from contextlib import asynccontextmanager
from collections.abc import AsyncIterator
from dataclasses import dataclass

from mcp.server.fastmcp import FastMCP
from mcp.server.transport_security import TransportSecuritySettings
from starlette.requests import Request
from starlette.responses import JSONResponse

from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE as SQL_RESPONSE_TYPE
from mindsdb.interfaces.storage import db
from mindsdb.utilities import log

logger = log.getLogger(__name__)


def _get_transport_security() -> TransportSecuritySettings:
    default_hosts = ["localhost:*", "127.0.0.1:*"]
    env_hosts = os.environ.get("MINDSDB_MCP_ALLOWED_HOSTS", "")
    if env_hosts:
        custom_hosts = [h.strip() for h in env_hosts.split(",") if h.strip()]
        for host in custom_hosts:
            if ":" not in host:
                default_hosts.append(f"{host}:*")
            default_hosts.append(host)
        logger.info(f"MCP transport security allowed hosts: {default_hosts}")
    return TransportSecuritySettings(allowed_hosts=default_hosts)


@dataclass
class AppContext:
    db: Any


@asynccontextmanager
async def app_lifespan(server: FastMCP) -> AsyncIterator[AppContext]:
    """Manage application lifecycle with type-safe context"""
    # Initialize on startup
    db.init()
    try:
        yield AppContext(db=db)
    finally:
        # TODO: We need better way to handle this in storage/db.py
        pass


# Configure server with lifespan and transport security
mcp = FastMCP(
    "MindsDB",
    lifespan=app_lifespan,
    dependencies=["mindsdb"],
    transport_security=_get_transport_security(),
)


# MCP Queries
LISTING_QUERY = "SHOW DATABASES"


query_tool_description = dedent("""\
    Executes a SQL query against MindsDB.

    A database must be specified either in the `context` parameter or directly in the query string (e.g., `SELECT * FROM my_database.my_table`). Queries like `SELECT * FROM my_table` will fail without a `context`.

    Args:
        query (str): The SQL query to execute.
        context (dict, optional): The default database context. For example, `{"db": "my_postgres"}`.

    Returns:
        A dictionary describing the result.
        - For a successful query with no data to return (e.g., an `UPDATE` statement), the response is `{"type": "ok"}`.
        - If the query returns tabular data, the response is a dictionary containing `data` (a list of rows) and `column_names` (a list of column names). For example: `{"type": "table", "data": [[1, "a"], [2, "b"]], "column_names": ["column_a", "column_b"]}`.
        - In case of an error, a response is `{"type": "error", "error_message": "the error message"}`.
""")


@mcp.tool(name="query", description=query_tool_description)
def query(query: str, context: dict | None = None) -> dict[str, Any]:
    """Execute a SQL query against MindsDB

    Args:
        query: The SQL query to execute
        context: Optional context parameters for the query

    Returns:
        Dict containing the query results or error information
    """

    if context is None:
        context = {}

    logger.debug(f"Incoming MCP query: {query}")

    mysql_proxy = FakeMysqlProxy()
    mysql_proxy.set_context(context)

    try:
        result = mysql_proxy.process_query(query)

        if result.type == SQL_RESPONSE_TYPE.OK:
            return {"type": SQL_RESPONSE_TYPE.OK}

        if result.type == SQL_RESPONSE_TYPE.TABLE:
            return {
                "type": SQL_RESPONSE_TYPE.TABLE,
                "data": result.result_set.to_lists(json_types=True),
                "column_names": [column.alias or column.name for column in result.result_set.columns],
            }
        else:
            return {"type": SQL_RESPONSE_TYPE.ERROR, "error_code": 0, "error_message": "Unknown response type"}

    except Exception as e:
        logger.exception("Error processing query:")
        return {"type": SQL_RESPONSE_TYPE.ERROR, "error_code": 0, "error_message": str(e)}


list_databases_tool_description = (
    "Returns a list of all database connections currently available in MindsDB. "
    + "The tool takes no parameters and responds with a list of database names, "
    + 'for example: ["my_postgres", "my_mysql", "test_db"].'
)


@mcp.tool(name="list_databases", description=list_databases_tool_description)
def list_databases() -> list[str]:
    """
    List all databases in MindsDB

    Returns:
       list[str]: list of databases
    """

    mysql_proxy = FakeMysqlProxy()

    try:
        result = mysql_proxy.process_query(LISTING_QUERY)
        if result.type == SQL_RESPONSE_TYPE.ERROR:
            return {
                "type": "error",
                "error_code": result.error_code,
                "error_message": result.error_message,
            }

        elif result.type == SQL_RESPONSE_TYPE.OK:
            return {"type": "ok"}

        elif result.type == SQL_RESPONSE_TYPE.TABLE:
            data = result.result_set.to_lists(json_types=True)
            data = [val[0] for val in data]
            return data

    except Exception as e:
        logger.exception("Error while retrieving list of databases")
        return {
            "type": "error",
            "error_code": 0,
            "error_message": str(e),
        }


def _get_status(request: Request) -> JSONResponse:
    """
    Status endpoint that returns basic server information.
    This endpoint can be used by the frontend to check if the MCP server is running.
    """

    status_info = {
        "status": "ok",
        "service": "mindsdb-mcp",
    }

    return JSONResponse(status_info)


def get_mcp_app():
    app = mcp.sse_app()
    app.add_route("/status", _get_status, methods=["GET"])
    return app



from typing import Optional, List

from sqlalchemy import null, func

import mindsdb.interfaces.storage.db as db
from mindsdb.utilities.context import context as ctx
import mindsdb.utilities.profiler as profiler


class PredictorRecordNotFound(Exception):
    def __init__(self, **kwargs):
        name = kwargs.get("name") or "-"
        predictor_id = kwargs.get("id") or "-"
        super().__init__(f"Predictor not found: name='{name}' id='{predictor_id}'")


class MultiplePredictorRecordsFound(Exception):
    def __init__(self, **kwargs):
        name = kwargs.get("name") or "-"
        predictor_id = kwargs.get("id") or "-"
        super().__init__(f"Found multiple predictor with: name='{name}' id='{predictor_id}'")


@profiler.profile()
def get_integration_record(name: str) -> db.Integration:
    company_id = ctx.company_id
    query = db.session.query(db.Integration).filter_by(company_id=company_id, name=name)
    if ctx.enforce_user_id:
        query = query.filter(db.Integration.user_id == ctx.user_id)
    record = query.first()
    return record


@profiler.profile()
def get_project_record(name: str) -> db.Project:
    company_id = ctx.company_id
    filters = [
        (func.lower(db.Project.name) == name),
        (db.Project.company_id == company_id),
        (db.Project.deleted_at == null()),
    ]
    if ctx.enforce_user_id:
        filters.append(db.Project.user_id == ctx.user_id)

    project_record = db.session.query(db.Project).filter(*filters).first()
    return project_record


@profiler.profile()
def get_project_records() -> List[db.Project]:
    company_id = ctx.company_id
    filters = [(db.Project.company_id == company_id), (db.Project.deleted_at == null())]
    if ctx.enforce_user_id:
        filters.append(db.Project.user_id == ctx.user_id)
    return db.session.query(db.Project).filter(*filters).all()


@profiler.profile()
def get_predictor_integration(record: db.Predictor) -> db.Integration:
    integration_record = db.session.query(db.Integration).filter_by(id=record.integration_id).first()
    return integration_record


@profiler.profile()
def get_predictor_project(record: db.Predictor) -> db.Project:
    project_record = db.session.query(db.Project).filter_by(id=record.project_id).first()
    return project_record


@profiler.profile()
def get_model_records(
    integration_id=None,
    active=True,
    deleted_at=null(),
    project_name: Optional[str] = None,
    ml_handler_name: Optional[str] = None,
    **kwargs,
):
    kwargs["company_id"] = ctx.company_id
    if ctx.enforce_user_id:
        kwargs["user_id"] = ctx.user_id

    if deleted_at is not None:
        kwargs["deleted_at"] = deleted_at
    if active is not None:
        kwargs["active"] = active

    if project_name is not None:
        project_record = get_project_record(name=project_name)
        if project_record is None:
            return []
        kwargs["project_id"] = project_record.id

    if ml_handler_name is not None:
        ml_handler_record = get_integration_record(name=ml_handler_name)
        if ml_handler_record is None:
            # raise Exception(f'unknown ml handler: {ml_handler_name}')
            return []
        kwargs["integration_id"] = ml_handler_record.id

    if integration_id is not None:
        kwargs["integration_id"] = integration_id

    return db.session.query(db.Predictor).filter_by(**kwargs).all()


@profiler.profile()
def get_model_record(
    except_absent=False,
    ml_handler_name: Optional[str] = None,
    project_name: Optional[str] = None,
    active: bool = True,
    deleted_at=null(),
    version: Optional[int] = None,
    **kwargs,
):
    kwargs["company_id"] = ctx.company_id
    if ctx.enforce_user_id:
        kwargs["user_id"] = ctx.user_id

    kwargs["deleted_at"] = deleted_at
    if active is not None:
        kwargs["active"] = active
    if version is not None:
        kwargs["version"] = version

    if project_name is not None:
        project_record = get_project_record(name=project_name)
        if project_record is None:
            return None
        kwargs["project_id"] = project_record.id

    if ml_handler_name is not None:
        ml_handler_record = get_integration_record(name=ml_handler_name)
        if ml_handler_record is None:
            # raise Exception(f'unknown ml handler: {ml_handler_name}')
            return []
        kwargs["integration_id"] = ml_handler_record.id

    records = db.session.query(db.Predictor).filter_by(**kwargs).all()
    if len(records) > 1:
        raise MultiplePredictorRecordsFound(**kwargs)
    if len(records) == 0:
        if except_absent is True:
            raise PredictorRecordNotFound(**kwargs)
        else:
            return None
    return records[0]



import re
import math
import logging
import collections
import json
from typing import List, Any, Optional, Dict, Tuple, Union, Callable

from pydantic import BaseModel, Field

from mindsdb.integrations.utilities.rag.retrievers.base import BaseRetriever, RunnableRetriever
from mindsdb.interfaces.knowledge_base.preprocessing.document_types import SimpleDocument

from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse
from mindsdb.integrations.libs.vectordatabase_handler import (
    DistanceFunction,
    VectorStoreHandler,
)
from mindsdb.integrations.utilities.rag.settings import (
    DatabaseSchema,
    TableSchema,
    ColumnSchema,
    ValueSchema,
    SearchKwargs,
)
from mindsdb.utilities import log

import numpy as np

logger = log.getLogger(__name__)


class MetadataFilter(BaseModel):
    """Represents an LLM generated metadata filter to apply to a PostgreSQL query."""

    attribute: str = Field(description="Database column to apply filter to")
    comparator: str = Field(description="PostgreSQL comparator to use to filter database column")
    value: Any = Field(description="Value to use to filter database column")


class AblativeMetadataFilter(MetadataFilter):
    """Adds additional fields to support ablation."""

    schema_table: str = Field(description="schema name of the table for this filter")
    schema_column: str = Field(description="schema name of the column for this filter")
    schema_value: str = Field(description="schema name of the value for this filter")


class MetadataFilters(BaseModel):
    """List of LLM generated metadata filters to apply to a PostgreSQL query."""

    filters: List[MetadataFilter] = Field(description="List of PostgreSQL metadata filters to apply for user query")


class SQLRetriever(BaseRetriever):
    """Retriever that uses a LLM to generate pgvector queries to do similarity search with metadata filters.

    How it works:

    1. Use a LLM to rewrite the user input to something more suitable for retrieval. For example:
    "Show me documents containing how to finetune a LLM please" --> "how to finetune a LLM"

    2. Use a LLM to generate structured metadata filters based on the user input. Provided
    metadata schemas & examples are used as additional context.

    3. Generate a prepared PostgreSQL query from the structured metadata filters.

    4. Actually execute the query against our vector database to retrieve documents & return them.
    """

    fallback_retriever: Any  # Must have get_relevant_documents or invoke method
    vector_store_handler: VectorStoreHandler
    # search parameters
    max_filters: int
    filter_threshold: float
    min_k: int

    # Schema description
    database_schema: Optional[DatabaseSchema] = None

    # Embeddings
    embeddings_model: Any  # Must have embed_query method
    search_kwargs: SearchKwargs

    # prompt templates
    rewrite_prompt_template: str

    # schema templates
    table_prompt_template: str
    column_prompt_template: str
    value_prompt_template: str

    # formatting templates
    boolean_system_prompt: str
    generative_system_prompt: str

    # SQL search config
    num_retries: int
    embeddings_table: str
    source_table: str
    source_id_column: str
    distance_function: DistanceFunction

    # Re-rank and metadata generation model.
    llm: Any  # Must have invoke method

    def _sort_schema_by_priority_key(
        self,
        schema_dict_item: Tuple[str, Union[TableSchema, ColumnSchema, ValueSchema]],
    ):
        return schema_dict_item[1].priority

    def _sort_schema_by_relevance_key(
        self,
        schema_dict_item: Tuple[str, Union[TableSchema, ColumnSchema, ValueSchema]],
    ):
        if schema_dict_item[1].relevance is not None:
            return schema_dict_item[1].relevance
        else:
            return 0

    def _sort_schema_by_key(
        self,
        schema: Union[DatabaseSchema, TableSchema, ColumnSchema],
        key: Callable,
        update: Dict[str, Any] = None,
    ) -> Union[DatabaseSchema, TableSchema, ColumnSchema]:
        """Takes a schema and converts its dict into an OrderedDict"""
        if isinstance(schema, DatabaseSchema):
            collection_key = "tables"
        elif isinstance(schema, TableSchema):
            collection_key = "columns"
        elif isinstance(schema, ColumnSchema):
            collection_key = "values"
        else:
            raise Exception("schema must be either a DatabaseSchema, TableSchema, or ColumnSchema.")

        if update is not None:
            ordered = collections.OrderedDict(sorted(update.items(), key=key, reverse=True))
        else:
            ordered = collections.OrderedDict(sorted(getattr(schema, collection_key).items(), key=key, reverse=True))
        schema = schema.model_copy(update={collection_key: ordered})

        return schema

    def _sort_database_schema_by_key(self, database_schema: DatabaseSchema, key: Callable) -> DatabaseSchema:
        """Re-build schema with OrderedDicts"""
        tables = {}
        # build new tables dict
        for table_key, table_schema in database_schema.tables.items():
            columns = {}
            # build new column dict
            for column_key, column_schema in table_schema.columns.items():
                # sort values directly and update column schema
                columns[column_key] = self._sort_schema_by_key(schema=column_schema, key=key)
            # update table schema and sort
            tables[table_key] = self._sort_schema_by_key(schema=table_schema, key=key, update=columns)
        # update table schema and sort
        database_schema = self._sort_schema_by_key(schema=database_schema, key=key, update=tables)

        return database_schema

    def _prepare_value_prompt(
        self,
        value_schema: ValueSchema,
        column_schema: ColumnSchema,
        table_schema: TableSchema,
        boolean_system_prompt: bool = True,
        format_instructions: Optional[str] = None,
    ) -> str:
        if boolean_system_prompt is True:
            system_prompt = self.boolean_system_prompt
        else:
            system_prompt = self.generative_system_prompt

        prepared_column_prompt = self._prepare_column_prompt(column_schema=column_schema, table_schema=table_schema)
        # Extract column schema string from prepared prompt (it's now a string)
        column_schema_str = (
            prepared_column_prompt.split("Query:")[0] if "Query:" in prepared_column_prompt else prepared_column_prompt
        )

        value_str = ""
        header_str = ""
        if type(value_schema.value) in [str, int, float, bool]:
            header_str = f"This schema describes a single value in the {column_schema.column} column."

            value_str = f"""
 -**Value**: {value_schema.value}
"""

        elif type(value_schema.value) is dict:
            header_str = f"This schema describes enumerated values in the {column_schema.column} column."

            value_str = """
## **Enumerated Values**

The values in the column are an enumeration of named values. These are listed below with format **[Column Value]**: [named value].
"""
            for value, value_name in value_schema.value.items():
                value_str += f"""
- **{value}:** {value_name}"""

        elif type(value_schema.value) is list:
            header_str = f"This schema describes some of the values in the {column_schema.column} column."

            value_str = """
## **Sample Values**

There are too many values in this column to list exhaustively. Below is a sampling of values found in the column:
"""
            for value in value_schema.value:
                value_str += f"""
- {value}"""

        if getattr(value_schema, "comparator", None) is not None:
            comparator_str = """

## **Comparators**

Below is a list of comparison operators for constructing filters for this value schema:
"""
            if type(value_schema.comparator) is str:
                comparator_str += f"""- {value_schema.comparator}
"""
            else:
                for comp in value_schema.comparator:
                    comparator_str += f"""- {comp}
"""
        else:
            comparator_str = ""

        if getattr(value_schema, "example_questions", None) is not None:
            example_str = """## **Example Questions**
"""
            for i, example in enumerate(value_schema.example_questions):
                example_str += f"""{i}. **Query:** {example.input} **Answer:** {example.output}
"""
        else:
            example_str = ""

        # Format prompt as string instead of ChatPromptTemplate
        format_instructions_str = format_instructions or ""
        prompt = f"""{system_prompt}

{self.value_prompt_template}

Format Instructions:
{format_instructions_str}

Header:
{header_str}

Column Schema:
{column_schema_str}

Value:
{value_str}

Comparator:
{comparator_str}

Type: {value_schema.type}
Description: {value_schema.description}
Usage: {value_schema.usage}

Examples:
{example_str}

Query: {{query}}"""
        return prompt

    def _prepare_column_prompt(
        self,
        column_schema: ColumnSchema,
        table_schema: TableSchema,
        boolean_system_prompt: bool = True,
    ) -> str:
        if boolean_system_prompt is True:
            system_prompt = self.boolean_system_prompt
        else:
            system_prompt = self.generative_system_prompt

        prepared_table_prompt = self._prepare_table_prompt(
            table_schema=table_schema, boolean_system_prompt=boolean_system_prompt
        )
        # Extract table schema string from prepared prompt (it's now a string)
        table_schema_str = (
            prepared_table_prompt.split("Query:")[0] if "Query:" in prepared_table_prompt else prepared_table_prompt
        )

        header_str = f"This schema describes a column in the {table_schema.table} table."

        value_str = """
## **Content**

Below is a description of the contents in this column in list format:
"""
        for value_schema in column_schema.values.values():
            value_str += f"""
- {value_schema.description}
"""
        value_str += """
**Important:** The above descriptions are not the actual values stored in this column. See the Value schema for actual values.
"""

        if getattr(column_schema, "examples", None) is not None:
            example_str = """## **Example Questions**
"""
            for example in column_schema.examples:
                example_str += f"""- {example}
"""
        else:
            example_str = ""

        # Format prompt as string instead of ChatPromptTemplate
        prompt = f"""{system_prompt}

{self.column_prompt_template}

Header:
{header_str}

Table Schema:
{table_schema_str}

Column: {column_schema.column}
Type: {column_schema.type}
Description: {column_schema.description}
Usage: {column_schema.usage}

Values:
{value_str}

Examples:
{example_str}

Query: {{query}}"""
        return prompt

    def _prepare_table_prompt(self, table_schema: TableSchema, boolean_system_prompt: bool = True) -> str:
        if boolean_system_prompt is True:
            system_prompt = self.boolean_system_prompt
        else:
            system_prompt = self.generative_system_prompt

        header_str = "This schema describes a table in the database."

        columns_str = ""
        for column_key, column_schema in table_schema.columns.items():
            columns_str += f"""
- **{column_schema.column}:** {column_schema.description}
"""

        if getattr(table_schema, "examples", None) is not None:
            example_str = """## **Example Questions**
"""
            for example in table_schema.examples:
                example_str += f"""- {example}
"""
        else:
            example_str = ""

        # Format prompt as string instead of ChatPromptTemplate
        prompt = f"""{system_prompt}

{self.table_prompt_template}

Header:
{header_str}

Table: {table_schema.table}
Description: {table_schema.description}
Usage: {table_schema.usage}

Columns:
{columns_str}

Examples:
{example_str}

Query: {{query}}"""
        return prompt

    def _rank_schema(self, prompt: str, query: str) -> float:
        """
        Rank schema by calling LLM with prompt and query.

        Args:
            prompt: Prompt template string with {query} placeholder
            query: Query string

        Returns:
            Relevance score between 0 and 1
        """
        # Format prompt with query
        formatted_prompt = prompt.format(query=query)

        try:
            # Call LLM - try to get logprobs if supported
            if hasattr(self.llm, "bind") and hasattr(self.llm.bind(logprobs=True), "invoke"):
                llm_with_logprobs = self.llm.bind(logprobs=True)
                output = llm_with_logprobs.invoke(formatted_prompt)
            else:
                # Fallback to regular invoke
                output = self.llm.invoke(formatted_prompt)

            # Try to extract logprobs from response
            score = None
            if hasattr(output, "response_metadata") and "logprobs" in output.response_metadata:
                logprobs = output.response_metadata["logprobs"]
                if "content" in logprobs:
                    for content in logprobs["content"]:
                        token = content.get("token", "").lower().strip()
                        logprob = content.get("logprob", 0.0)
                        if token == "yes":
                            score = (1 + math.exp(logprob)) / 2
                            break
                        elif token == "no":
                            score = (1 - math.exp(logprob)) / 2
                            break

            # If no logprobs, try to parse yes/no from content
            if score is None:
                content_text = ""
                if hasattr(output, "content"):
                    content_text = output.content.lower().strip()
                elif isinstance(output, str):
                    content_text = output.lower().strip()
                else:
                    content_text = str(output).lower().strip()

                if "yes" in content_text:
                    score = 0.75  # Default positive score
                elif "no" in content_text:
                    score = 0.25  # Default negative score
                else:
                    score = 0.5  # Neutral score

            if score is None:
                score = 0.0

        except Exception as e:
            logger.warning(f"Error ranking schema: {e}")
            score = 0.0

        return score

    def _breadth_first_search(self, query: str, greedy: bool = False) -> Tuple:
        """Search breadth wise through Tables, then Columns, then Values.Uses a greedy strategy to maximize quota if greedy=True, otherwise a dynamic strategy."""

        # sort based on priority
        ordered_database_schema = self._sort_database_schema_by_key(
            database_schema=self.database_schema, key=self._sort_schema_by_priority_key
        )

        #  Rank Tables ########################################################
        greedy_count = 0
        tables = {}
        # rank tables by relevance
        for table_key, table_schema in ordered_database_schema.tables.items():
            prompt: str = self._prepare_table_prompt(table_schema=table_schema, boolean_system_prompt=True)
            table_schema.relevance = self._rank_schema(prompt=prompt, query=query)

            # only keep greedy tables
            tables[table_key] = table_schema

            if greedy:
                if table_schema.relevance >= ordered_database_schema.filter_threshold:
                    greedy_count += 1
                if greedy_count >= ordered_database_schema.max_filters:
                    break

        #  sort tables
        ordered_database_schema = self._sort_schema_by_key(
            schema=ordered_database_schema,
            key=self._sort_schema_by_relevance_key,
            update=tables,
        )

        #  Rank Columns #######################################################
        #  iterate through tables to rank columns
        tables = {}
        table_count = 0  # take only the top n number of tables specified by the databases max filters
        for table_key, table_schema in ordered_database_schema.tables.items():
            # only drop into tables above the filter threshold
            if table_schema.relevance >= ordered_database_schema.filter_threshold:
                greedy_count = 0
                # rank columns by relevance
                columns = {}
                for column_key, column_schema in table_schema.columns.items():
                    prompt: str = self._prepare_column_prompt(
                        column_schema=column_schema,
                        table_schema=table_schema,
                        boolean_system_prompt=True,
                    )
                    column_schema.relevance = self._rank_schema(prompt=prompt, query=query)

                    columns[column_key] = column_schema

                    if greedy:
                        if column_schema.relevance >= table_schema.filter_threshold:
                            greedy_count += 1
                        if greedy_count >= table_schema.max_filters:
                            break

                # sort columns and keep only columns that made the cut.
                tables[table_key] = self._sort_schema_by_key(
                    table_schema, key=self._sort_schema_by_relevance_key, update=columns
                )

                table_count += 1
                if table_count >= ordered_database_schema.max_filters:
                    break

        # sort tables and keep only tables that made the cut.
        ordered_database_schema = self._sort_schema_by_key(
            ordered_database_schema,
            key=self._sort_schema_by_relevance_key,
            update=tables,
        )

        #  Rank Values ########################################################
        #  iterate through tables to rank values
        tables = {}
        for table_key, table_schema in ordered_database_schema.tables.items():
            columns = {}
            column_count = 0
            # iterate through columns to rank values
            for column_key, column_schema in table_schema.columns.items():
                if column_schema.relevance >= table_schema.filter_threshold:
                    greedy_count = 0
                    values = {}
                    #  rank values by relevance
                    for value_key, value_schema in column_schema.values.items():
                        prompt: str = self._prepare_value_prompt(
                            value_schema=value_schema,
                            column_schema=column_schema,
                            table_schema=table_schema,
                            boolean_system_prompt=True,
                        )
                        value_schema.relevance = self._rank_schema(prompt=prompt, query=query)

                        values[value_key] = value_schema

                        if greedy:
                            if value_schema.relevance >= column_schema.filter_threshold:
                                greedy_count += 1
                            if greedy_count >= column_schema.max_filters:
                                break

                    # sort values and keep only values that make the cut
                    columns[column_key] = self._sort_schema_by_key(
                        column_schema,
                        key=self._sort_schema_by_relevance_key,
                        update=values,
                    )

                    column_count += 1
                    if column_count >= table_schema.max_filters:
                        break

            # sort columns and keep only columns that made the cut
            tables[table_key] = self._sort_schema_by_key(
                table_schema, key=self._sort_schema_by_relevance_key, update=columns
            )

        # sort tables and keep only tables that made the cut.
        ordered_database_schema = self._sort_schema_by_key(
            ordered_database_schema,
            key=self._sort_schema_by_relevance_key,
            update=tables,
        )

        #  discard low ranked values ###################################################################################
        tables = {}
        for table_key, table_schema in ordered_database_schema.tables.items():
            columns = {}
            # iterate through columns to rank values
            for column_key, column_schema in table_schema.columns.items():
                value_count = 0
                values = {}
                #  rank values by relevance
                for value_key, value_schema in column_schema.values.items():
                    if value_schema.relevance >= column_schema.filter_threshold:
                        values[value_key] = value_schema

                        value_count += 1
                        if value_count >= column_schema.max_filters:
                            break

                # sort values and keep only values that make the cut
                columns[column_key] = self._sort_schema_by_key(
                    column_schema,
                    key=self._sort_schema_by_relevance_key,
                    update=values,
                )

            # sort columns and keep only columns that made the cut
            tables[table_key] = self._sort_schema_by_key(
                table_schema, key=self._sort_schema_by_relevance_key, update=columns
            )

        # sort tables and keep only tables that made the cut.
        ordered_database_schema = self._sort_schema_by_key(
            ordered_database_schema,
            key=self._sort_schema_by_relevance_key,
            update=tables,
        )

        ranked_database_schema = ordered_database_schema

        #  Build Ablation #####################################################

        ablation_value_dict = {}
        # assemble a relevance dictionary
        for table_key, table_schema in ordered_database_schema.tables.items():
            for column_key, column_schema in table_schema.columns.items():
                for value_key, value_schema in column_schema.values.items():
                    ablation_value_dict[(table_key, column_key, value_key)] = value_schema.relevance

        ablation_value_dict = collections.OrderedDict(sorted(ablation_value_dict.items(), key=lambda x: x[1]))

        relevance_scores = list(ablation_value_dict.values())
        if len(relevance_scores) > 0:
            ablation_quantiles = np.quantile(relevance_scores, np.linspace(0, 1, self.num_retries + 2)[1:-1])
        else:
            ablation_quantiles = None

        return ranked_database_schema, ablation_value_dict, ablation_quantiles

    def _dynamic_ablation(
        self,
        metadata_filters: List[AblativeMetadataFilter],
        ablation_value_dict,
        ablation_quantiles,
        retry: int,
    ):
        """Ablate metadata filters in aggregate by quantiles until the required minimum number of documents are returned."""

        ablated_dict = {}
        for key, value in ablation_value_dict.items():
            if value >= ablation_quantiles[retry]:
                ablated_dict[key] = value

        #  discard low ranked filters ##################################################################################
        ablated_filters = []
        for filter in metadata_filters:
            for key in ablated_dict.keys():
                if filter.schema_table in key and filter.schema_column in key and filter.schema_value in key:
                    ablated_filters.append(filter)

        return ablated_filters

    def depth_first_search(self, greedy=True):
        """Search depth wise through Tables, then Columns, then Values. Uses a greedy strategy to maximize quota if greedy=True, otherwise a dynamic strategy."""
        pass

    def depth_first_ablation(self):
        """Ablate metadata filters in reverse depth first search until the required minimum number of documents are returned."""
        pass

    def _prepare_retrieval_query(self, query: str) -> str:
        """Rewrite query to be suitable for retrieval using LLM"""
        # Format prompt with query
        formatted_prompt = self.rewrite_prompt_template.format(input=query)

        # Call LLM
        llm_response = self.llm.invoke(formatted_prompt)

        # Extract content from LLM response
        if hasattr(llm_response, "content"):
            return llm_response.content
        elif isinstance(llm_response, str):
            return llm_response
        else:
            return str(llm_response)

    def _prepare_pgvector_query(
        self,
        ranked_database_schema: DatabaseSchema,
        metadata_filters: List[AblativeMetadataFilter],
        retry: int = 0,
    ) -> str:
        # Base select JOINed with document source table.
        base_query = f"""SELECT * FROM {self.embeddings_table} AS e INNER JOIN {self.source_table} AS s ON (e.metadata->>'original_row_id')::int = s."{self.source_id_column}" """

        # return an empty string if schema has not been ranked
        if not ranked_database_schema:
            return ""

        # Add Table JOIN statements
        join_clauses = set()
        for metadata_filter in metadata_filters:
            join_clause = ranked_database_schema.tables[metadata_filter.schema_table].join
            if join_clause in join_clauses:
                continue
            else:
                join_clauses.add(join_clause)
                base_query += join_clause + " "

        # Add WHERE conditions from metadata filters
        if metadata_filters:
            base_query += "WHERE "
        for i, filter in enumerate(metadata_filters):
            value = filter.value
            if isinstance(value, str):
                value = f"'{value}'"
            base_query += f'"{filter.attribute}" {filter.comparator} {value}'
            if i < len(metadata_filters) - 1:
                base_query += " AND "

        base_query += (
            f" ORDER BY e.embeddings {self.distance_function.value[0]} '{{embeddings}}' LIMIT {self.search_kwargs.k};"
        )
        return base_query

    def _generate_filter(self, prompt: str, query: str) -> MetadataFilter:
        """Generate metadata filter using LLM"""
        # Format prompt with query
        formatted_prompt = prompt.format(query=query)

        # Call LLM
        llm_response = self.llm.invoke(formatted_prompt)

        # Extract content from LLM response
        if hasattr(llm_response, "content"):
            response_text = llm_response.content
        elif isinstance(llm_response, str):
            response_text = llm_response
        else:
            response_text = str(llm_response)

        # Parse JSON response to get MetadataFilter
        try:
            parsed = json.loads(response_text)
            # If it's a dict, try to create MetadataFilter
            if isinstance(parsed, dict):
                return MetadataFilter(**parsed)
            else:
                # If it's already a MetadataFilter-like object
                return parsed
        except (json.JSONDecodeError, TypeError, Exception) as e:
            logger.warning(f"Error parsing filter response: {e}")
            # Return empty filter on error
            return MetadataFilter(attribute="", comparator="=", value="")

    def _generate_metadata_filters(
        self, query: str, ranked_database_schema
    ) -> Union[List[AblativeMetadataFilter], HandlerResponse]:
        """Generate metadata filters using LLM"""

        metadata_filter_list = []
        #  iterate through tables to rank values
        for table_key, table_schema in ranked_database_schema.tables.items():
            # iterate through columns to rank values
            for column_key, column_schema in table_schema.columns.items():
                if column_schema.relevance >= table_schema.filter_threshold:
                    #  generate filters
                    for value_key, value_schema in column_schema.values.items():
                        # must use generation if field is a dictionary of tuples or a list
                        if type(value_schema.value) in [list, dict]:
                            try:
                                # Create format instructions for JSON output
                                format_instructions = """Return a JSON object with the following structure:
{
  "attribute": "column_name",
  "comparator": "comparison_operator",
  "value": "filter_value"
}"""

                                metadata_prompt: str = self._prepare_value_prompt(
                                    format_instructions=format_instructions,
                                    value_schema=value_schema,
                                    column_schema=column_schema,
                                    table_schema=table_schema,
                                    boolean_system_prompt=False,
                                )

                                # Call LLM directly
                                formatted_prompt = metadata_prompt.format(query=query)
                                llm_response = self.llm.invoke(formatted_prompt)

                                # Extract content from LLM response
                                if hasattr(llm_response, "content"):
                                    metadata_filter_output = llm_response.content
                                elif isinstance(llm_response, str):
                                    metadata_filter_output = llm_response
                                else:
                                    metadata_filter_output = str(llm_response)

                                # If the LLM outputs raw JSON, use it as-is.
                                # If the LLM outputs anything including a json markdown section, use the last one.
                                json_markdown_output = re.findall(r"```json.*?```", metadata_filter_output, re.DOTALL)
                                if json_markdown_output:
                                    metadata_filter_output = json_markdown_output[-1]
                                    # Clean the json tags.
                                    metadata_filter_output = metadata_filter_output[7:]
                                    metadata_filter_output = metadata_filter_output[:-3]

                                # Parse JSON directly instead of using PydanticOutputParser
                                parsed = json.loads(metadata_filter_output.strip())
                                model_dump = {
                                    "attribute": parsed.get("attribute", ""),
                                    "comparator": parsed.get("comparator", "="),
                                    "value": parsed.get("value", ""),
                                    "schema_table": table_key,
                                    "schema_column": column_key,
                                    "schema_value": value_key,
                                }
                                metadata_filter = AblativeMetadataFilter(**model_dump)
                            except (json.JSONDecodeError, TypeError, Exception) as e:
                                logger.warning(
                                    f"LLM failed to generate structured metadata filters: {e}",
                                    exc_info=logger.isEnabledFor(logging.DEBUG),
                                )
                                return HandlerResponse(RESPONSE_TYPE.ERROR, error_message=str(e))
                        else:
                            metadata_filter = AblativeMetadataFilter(
                                attribute=column_schema.column,
                                comparator=value_schema.comparator,
                                value=value_schema.value,
                                schema_table=table_key,
                                schema_column=column_key,
                                schema_value=value_key,
                            )
                        metadata_filter_list.append(metadata_filter)

        return metadata_filter_list

    def _prepare_and_execute_query(
        self,
        ranked_database_schema: DatabaseSchema,
        metadata_filters: List[AblativeMetadataFilter],
        embeddings_str: str,
    ) -> HandlerResponse:
        try:
            checked_sql_query = self._prepare_pgvector_query(ranked_database_schema, metadata_filters)
            checked_sql_query_with_embeddings = checked_sql_query.format(embeddings=embeddings_str)
            return self.vector_store_handler.native_query(checked_sql_query_with_embeddings)
        except Exception as e:
            logger.warning(
                f"Failed to prepare and execute SQL query from structured metadata: {e}",
                exc_info=logger.isEnabledFor(logging.DEBUG),
            )
            return HandlerResponse(RESPONSE_TYPE.ERROR, error_message=str(e))

    def _get_relevant_documents(self, query: str, *, run_manager: Optional[Any] = None) -> List[Any]:
        # Rewrite query to be suitable for retrieval.
        retrieval_query = self._prepare_retrieval_query(query)

        # Embed the rewritten retrieval query & include it in the similarity search pgvector query.
        embedded_query = self.embeddings_model.embed_query(retrieval_query)

        # Search for relevant filters
        ranked_database_schema, ablation_value_dict, ablation_quantiles = self._breadth_first_search(query=query)

        # Generate metadata filters
        metadata_filters = self._generate_metadata_filters(query=query, ranked_database_schema=ranked_database_schema)

        if type(metadata_filters) is list:
            # Initial Execution of the similarity search with metadata filters.
            document_response = self._prepare_and_execute_query(
                ranked_database_schema=ranked_database_schema,
                metadata_filters=metadata_filters,
                embeddings_str=str(embedded_query),
            )
            num_retries = 0
            while num_retries < self.num_retries:
                if (
                    document_response.resp_type != RESPONSE_TYPE.ERROR
                    and len(document_response.data_frame) >= self.min_k
                ):
                    # Successfully retrieved k documents to send to re-ranker.
                    break
                elif document_response.resp_type == RESPONSE_TYPE.ERROR:
                    # LLMs won't always generate structured metadata so we should have a fallback after retrying.
                    logger.info(f"SQL Retriever query failed with error {document_response.error_message}")
                else:
                    logger.info(
                        f"SQL Retriever did not retrieve {self.min_k} documents: {len(document_response.data_frame)} documents retrieved."
                    )

                ablated_metadata_filters = self._dynamic_ablation(
                    metadata_filters=metadata_filters,
                    ablation_value_dict=ablation_value_dict,
                    ablation_quantiles=ablation_quantiles,
                    retry=num_retries,
                )

                document_response = self._prepare_and_execute_query(
                    ranked_database_schema=ranked_database_schema,
                    metadata_filters=ablated_metadata_filters,
                    embeddings_str=str(embedded_query),
                )

                num_retries += 1

            retrieved_documents = []
            if document_response.resp_type != RESPONSE_TYPE.ERROR:
                document_df = document_response.data_frame
                for _, document_row in document_df.iterrows():
                    retrieved_documents.append(
                        SimpleDocument(
                            page_content=document_row.get("content", ""),
                            metadata=document_row.get("metadata", {}),
                        )
                    )
            if retrieved_documents:
                return retrieved_documents

            # If the SQL query constructed did not return any documents, fallback.
            logger.info("No documents returned from SQL retriever, using fallback retriever.")
            return self._retrieve_from_fallback_retriever(retrieval_query)
        else:
            # If no metadata fields could be generated fallback.
            logger.info("No metadata fields were successfully generated, using fallback retriever.")
            return self._retrieve_from_fallback_retriever(retrieval_query)

    def _retrieve_from_fallback_retriever(self, query: str) -> List[Any]:
        """Retrieve documents from fallback retriever using duck typing"""
        if hasattr(self.fallback_retriever, "_get_relevant_documents"):
            return self.fallback_retriever._get_relevant_documents(query)
        elif hasattr(self.fallback_retriever, "get_relevant_documents"):
            return self.fallback_retriever.get_relevant_documents(query)
        elif hasattr(self.fallback_retriever, "invoke"):
            return self.fallback_retriever.invoke(query)
        else:
            raise ValueError(
                "Fallback retriever must have _get_relevant_documents, get_relevant_documents, or invoke method"
            )

    def invoke(self, query: str) -> List[Any]:
        """Sync invocation - retrieve documents for a query"""
        return self._get_relevant_documents(query)

    async def ainvoke(self, query: str) -> List[Any]:
        """Async invocation - retrieve documents for a query"""
        import asyncio

        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, self._get_relevant_documents, query)

    def get_relevant_documents(self, query: str) -> List[Any]:
        """Get relevant documents (sync)"""
        return self._get_relevant_documents(query)

    def as_runnable(self) -> RunnableRetriever:
        """Return self as a runnable retriever"""
        return self
